import BayesBoom.boom as boom
import BayesBoom.R as R

from abc import ABC, abstractmethod

HOLIDAY_REGISTRY = {}

# TODO update holiday registry by country.

days_in_month = (-1,
                 31,  # January
                 28,  # February
                 31,  # March
                 30,  # April
                 31,  # May
                 30,  # June
                 31,  # July
                 31,  # August
                 30,  # September
                 31,  # October
                 30,  # November
                 31,  # December
                 )


def register_holiday(name, holiday_object):
    global HOLIDAY_REGISTRY
    HOLIDAY_REGISTRY[name] = holiday_object


class Holiday(ABC):
    """
    """
    @abstractmethod
    def boom(self):
        """
        Return a boom holiday corresponding to the information in this holiday
        object.
        """

    @property
    def maximum_window_width(self):
        """
        Holidays exert influence before or after the day itself.  The
        maximum_window_width is the number of days of infludence a holiday has
        each time it occurs.
        """
        return self._boom_holiday.maximum_window_width

    def active(self, arbitrary_date):
        """
        Returns True iff 'arbitrary_date' falls inisde the influence window for
        the holiday.
        """
        return self._boom_holiday.active(R.to_boom_date(arbitrary_date))

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_boom_holiday"]

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._boom_holiday = self.boom()


class SingleDayHoliday(Holiday):
    """
    A SingleDayHoliday is a holiday associated with a specific date.  Its
    influence can extend beyond that date, but (e.g.) February 14 is
    Valentine's day.  Most Holidays are SingleDayHolidays, some religious
    holidays (e.g. Passover) and some sporting events (e.g. the Olympics or the
    World Cup) are not.
    """

    def date_on_or_after(self, date):
        """
        Either 'date' or the first instance of the holiday after 'date'.
        """
        ans = self._boom_holiday.date_on_or_after(R.to_boom_date(date))
        return R.to_pd_timestamp(ans)

    def date_on_or_before(self, date):
        """
        Either 'date' or the first instance of the holiday before 'date'.
        """
        return R.to_pd_timestamp(self._boom_holiday.date_on_or_before(
            R.to_boom_date(date)))

    def nearest(self, date):
        """
        The date of the closest instance of this holiday to 'date'.
        """
        return R.to_pd_timestamp(self._boom_holiday.nearest(
            R.to_boom_date(date)))


class OrdinaryAnnualHoliday(SingleDayHoliday):
    """
    An OrdinaryAnnualHoliday is a Holiday that occurs once per year, with a
    fixed-sized window of influence.  An OrdinaryAnnualHoliday keeps track of
    two integers: days_before and days_after, that define its influence window.
    """
    def __init__(self, days_before: int = 1, days_after: int = 1):
        if not days_before >= 0:
            raise Exception("'days_before' must be non-negative.")
        if not days_after >= 0:
            raise Exception("'days_after' must be non-negative.")
        self._days_before = int(days_before)
        self._days_after = int(days_after)

    def date(self, year):
        return R.to_pd_timestamp(self._boom_holiday.date(int(year)))


class FixedDateHoliday(OrdinaryAnnualHoliday):
    """
    A holiday that occurs on the same calendar date each year.
    """
    def __init__(self,
                 month: int,
                 day: int,
                 days_before: int = 1,
                 days_after: int = 1):
        super().__init__(days_before, days_after)
        if month < 1 or month > 12:
            raise Exception("'month' must be between 1 and 12.")
        if day < 1 or day > days_in_month[month]:
            raise Exception(
                f"For month {month}, the day argument must be between 1 "
                f"and {days_in_month[month]}."
            )
        self._month = int(month)
        self._day = int(day)
        self._boom_holiday = self.boom()

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.FixedDateHoliday(
            self._month,
            self._day,
            self._days_before,
            self._days_after)


class NthWeekdayInMonthHoliday(OrdinaryAnnualHoliday):
    """
    A holidy that occurs each year on the same day of the week, defined as the
    Nth such dat in a month.  For example, US Labor Day is the first Monday in
    September.
    """
    def __init__(self,
                 which_week: int,
                 day_of_week: int,
                 month: int,
                 days_before: int = 1,
                 days_after: int = 1):
        super().__init__(days_before, days_after)
        if which_week < 1 or which_week > 5:
            raise Exception("which_week must be between 1 and 5.")
        if month < 1 or month > 12:
            raise Exception("month must be between 1 and 12.")
        if day_of_week < 1 or day_of_week > 7:
            raise Exception("day_of_week must be between 1 and 7.")
        self._which_week = which_week
        self._day_of_week = day_of_week
        self._month = month
        self._boom_holiday = self.boom()

    def boom(self):
        # TODO(steve): Document the mapping between int's and days.
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.NthWeekdayInMonthHoliday(
            self._which_week,
            self._day_of_week,
            self._month,
            self._days_before,
            self._days_after)


class LastWeekdayInMonthHoliday(OrdinaryAnnualHoliday):
    """
    A holiday that occurs each year on the final weekday in a given month.  For
    example, US Memorial Day is the last Monday in May.
    """
    def __init__(self,
                 day_of_week: int,
                 month: int,
                 days_before: int = 1,
                 days_after: int = 1):
        super().__init__(days_before, days_after)
        if month < 1 or month > 12:
            raise Exception("month must be between 1 and 12")
        if day_of_week < 1 or day_of_week > 7:
            raise Exception("day_of_week must be between 1 and 7")
        self._month = int(month)
        self._day_of_week = int(day_of_week)
        self._boom_holiday = self.boom()

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.LastWeekdayInMonthHoliday(
            self._day_of_week,
            self._month,
            self._days_before,
            self._days_after)


class DateRangeHoliday(Holiday):
    """
    An irregular holiday, for which the influence window must be handled
    manually.
    """
    def __init__(self, start, end):
        self._start = start
        self._end = end
        self._boom_holiday = self.boom()

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        start_days = [R.to_boom_date(x) for x in self._start]
        end_days = [R.to_boom_date(x) for x in self._end]

        return boom.DateRangeHoliday(start_days, end_days)


# ======================================================================
# Some holidays that require smarts from BOOM.
# ======================================================================

class EasterSunday(OrdinaryAnnualHoliday):
    def __init__(self, days_before: int = 1, days_after: int = 1):
        self._days_before = int(days_before)
        self._days_after = int(days_after)
        self._boom_holiday = self.boom()

    def date(self, year):
        return R.to_pd_timestamp(self._boom_holiday.date(int(year)))

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.EasterSunday(
            self._days_before, self._days_after)


class USDaylightSavingsTimeBegins(OrdinaryAnnualHoliday):
    def __init__(self, days_before: int = 1, days_after: int = 1):
        self._days_before = int(days_before)
        self._days_after = int(days_after)
        self._boom_holiday = self.boom()

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.USDaylightSavingsTimeBegins(
            self._days_before, self._days_after)


class USDaylightSavingsTimeEnds(OrdinaryAnnualHoliday):
    def __init__(self, days_before: int = 1, days_after: int = 1):
        self._days_before = int(days_before)
        self._days_after = int(days_after)
        self._boom_holiday = self.boom()

    def boom(self):
        if hasattr(self, "_boom_holiday"):
            return self._boom_holiday
        return boom.USDaylightSavingsTimeEnds(
            self._days_before, self._days_after)


class HolidayFactory:

    def create_boom_holiday(self,
                            holiday,
                            days_before: int = 1,
                            days_after: int = 1):
        if isinstance(holiday, Holiday):
            return holiday.boom()
        elif isinstance(holiday, str) and holiday in HOLIDAY_REGISTRY.keys():
            if holiday in us_named_holidays:
                return boom.create_named_holiday(
                    holiday, int(days_before), int(days_after))
        else:
            raise Exception(f"Unrecognized holiday: {holiday}.")


us_named_holidays = [
    "NewYearsDay",
    "MartinLutherKingDay",
    "SuperBowlSunday",
    "PresidentsDay",
    "ValentinesDay",
    "SaintPatricksDay",
    "USDaylightSavingsTimeBegins",
    "USDaylightSavingsTimeEnds",
    "EasterSunday",
    "USMothersDay",
    "IndependenceDay",
    "LaborDay",
    "ColumbusDay",
    "Halloween",
    "Thanksgiving",
    "MemorialDay",
    "VeteransDay",
    "Christmas",
]



import unittest
import numpy as np
import pandas as pd
import pickle
import json

import pdb

import matplotlib.pyplot as plt

from BayesBoom.R import delete_if_present

from BayesBoom.bsts import (
    Bsts,
    AirPassengers,
    LocalLinearTrendStateModel,
    GeneralSeasonalLLT,
)


def default_colnames(n):
    return ["X" + str(i + 1) for i in range(n)]


class TestSeasonalLLT(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def tearDown(self):
        delete_if_present("seasonal_llt.pkl")

    def test_mcmc(self):
        model = Bsts()
        model.add_state(LocalLinearTrendStateModel(AirPassengers))
        model.add_state(GeneralSeasonalLLT(AirPassengers, nseasons=12))

        model.train(data=AirPassengers, niter=1000)

        fname = "seasonal_llt.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(model, pkl)

        with open(fname, "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertEqual(model.time_dimension, m2.time_dimension)
        self.assertIsInstance(m2, Bsts)

        fig = model.plot("comp")
        fig.show()

        pred = model.predict(12)
        fig2, ax = pred.plot()
        fig2.show()

    def test_plots(self):
        pass

    def test_predictions(self):
        pass

    def test_serialization(self):
        pass


_debug_mode = True


if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestSeasonalLLT()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_mcmc()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



from .bsts import Bsts, BstsPrediction, extend_timestamps, compare_bsts_models

from .state_models import StateModel
from .ar import ArStateModel, AutoArStateModel, SpikeSlabArPrior
from .dynamic_regression_state_model import DynamicRegressionStateModel
from .general_seasonal_llt import GeneralSeasonalLLT
from .local_level import LocalLevelStateModel
from .local_linear_trend import LocalLinearTrendStateModel
from .seasonal import SeasonalStateModel
from .semilocal_linear_trend import SemilocalLinearTrendStateModel
from .student_local_linear_trend import StudentLocalLinearTrendStateModel
from .trig import TrigStateModel

from .holiday import (
    Holiday,
    FixedDateHoliday,
    NthWeekdayInMonthHoliday,
    LastWeekdayInMonthHoliday,
    EasterSunday,
    USDaylightSavingsTimeBegins,
    USDaylightSavingsTimeEnds,
)

from .holiday_models import RegressionHolidayStateModel

from .data import AirPassengers

from .test_utilities import (  # noqa
    simulate_random_walk,
    simulate_student_random_walk,
    simulate_student_local_linear_trend,
    simulate_local_linear_trend,
)

__all__ = [
    "Bsts",
    "BstsPrediction",
    "StateModel",
    "ArStateModel",
    "AutoArStateModel",
    "DynamicRegressionStateModel",
    "EasterSunday",
    "FixedDateHoliday",
    "GeneralSeasonalLLT",
    "Holiday",
    "LastWeekdayInMonthHoliday",
    "LocalLevelStateModel",
    "LocalLinearTrendStateModel",
    "NthWeekdayInMonthHoliday",
    "RegressionHolidayStateModel",
    "SeasonalStateModel",
    "SemilocalLinearTrendStateModel",
    "SpikeSlabArPrior",
    "StudentLocalLinearTrendStateModel",
    "TrigStateModel",
    "USDaylightSavingsTimeBegins",
    "USDaylightSavingsTimeEnds",
    "AirPassengers",
    "compare_bsts_models",
    "extend_timestamps",
]



import numpy as np
from .plots import _skim_plot_options, _set_plot_options


class Density:
    """
    A Gaussian kernel density estimate of a scalar valued variable.
    """

    def __init__(self, x, bw: float = None):
        """
        Args:
          x: a 1-d numpy array or object convertible to such.
          bw:  None means to use R's default bandwidth.
        """
        from scipy.stats import gaussian_kde

        x = np.array(x).ravel()
        finite = np.isfinite(x)
        x = x[finite]
        x.sort()
        if bw is None:
            bw = self._default_bandwidth(x)

        self._kde = gaussian_kde(x)
        self._grid = np.linspace(x[0], x[-1], 100)
        self._density_values = self(self._grid)

    def plot(self, ax=None, **kwargs):
        """
        Plot the density function on the supplied set of Axes.
        """
        import matplotlib.pyplot as plt

        fig = None
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        plot_options, kwargs = _skim_plot_options(**kwargs)
        ax.plot(self._grid, self._density_values, **kwargs)
        _set_plot_options(ax, **plot_options)
        if fig is not None:
            fig.show()
        return ax

    def __call__(self, x):
        """
        Treat a density object as function to be evaluated at x.
        """
        return self._kde(x)

    @staticmethod
    def _default_bandwidth(x):
        """
        R's default logic for selecting the bandwidth of a kernel density
        estimator for x.  In R this function is called 'bw.nrd0'.
        """
        sdx = np.std(x, ddof=1)
        quantiles = np.quantile(x, (.25, .75))
        iqr = np.max(quantiles) - np.min(quantiles)

        hi = sdx
        lo = min(hi, iqr / 1.34)
        if lo <= 0.0:
            lo = hi
            if lo <= 0.0:
                lo = abs(x[0])
                if lo <= 0.0:
                    lo = 1
        return .9 * lo * len(x) ** -.2



import unittest
import os

from BayesBoom.spikeslab import mlogit_spike
from BayesBoom.test_utils import find_project_root

import numpy as np
import pandas as pd


class MlogitTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_mcmc(self):
        dir = os.path.join(find_project_root(), "BayesBoom", "spikeslab")
        fname = os.path.join(dir, "autopref.txt")
        data = pd.read_csv(fname)

        model = mlogit_spike(
            data["type"],
            "age + sex + married",
            {},
            niter=100,
            data=data)
        self.assertIsInstance(model, mlogit_spike)


_debug_mode = True

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = MlogitTest()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_mcmc()

    print("Goodbye, cruel world!")

if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd

# A collection of free utility functions helpful for mixture modeling.


def normalize_logprob(log_probs):
    """
    Convert a matrix of unnormalized log probabilities into normalized
    probabilities.

    Args:
      log_probs: A matrix where each row is a set of log probabilities to be
        exponentiated and normalized so the exponentiated row sums to one.
    """
    if isinstance(log_probs, (pd.DataFrame, pd.Series)):
        log_probs = log_probs.values

    row_max = log_probs.max(axis=1).reshape((-1, 1))
    probs = np.exp(log_probs - row_max)
    row_totals = probs.sum(axis=1).reshape((-1, 1))
    return probs / row_totals



# import numpy as np
# import bsts
# import matplotlib.pyplot as plt

# y = np.log(bsts.AirPassengers)

# model = bsts.Bsts()

# # model.add_state(bsts.LocalLevelStateModel(y))
# model.add_state(bsts.SemilocalLinearTrendStateModel(y))
# model.add_state(bsts.SeasonalStateModel(y, nseasons=12))
# model.train(data=y, niter=100)

# model.plot("comp")
# plt.show()



# import matplotlib.pyplot as plt
# import seaborn as sns
import BayesBoom.boom as boom
import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
import pickle
import time


class MixedDataImputer:
    """
    Imputes missing data in a pd.DataFrame containing a mix of numeric and
    categorical data.

    Example use:
    """

    def __init__(self):
        """
        Create an empty MixedDataImputer.
        """
        self._model = None
        self._numeric_colnames = None
        self._categorical_colnames = None

        # A dict keyed by variable names of numeric columns.  The values are
        # lists of numeric values to be treated as atoms for that variable.
        self._atoms = {}

        self._levels = {}

        # A dict keyed by variable names of numeric columns.  The values are
        # 1-d numpy arrays containing prior counts for the atom values for that
        # variable.  If there are 'k' atoms then the vector must be of length
        # 'k+1', with the final count corresponding to the implicit continuous
        # atom.
        self._atom_prior = {}

        # A dict keyed by variable names of numeric columns.  The values are
        # 1-d numpy arrays containing prior counts for the level values for
        # that variable.  If there are 'k' levels then the vector must be of
        # length 'k'.
        self._level_prior = {}

        self._dataset_encoder = None
        self.coefficient_draws = np.zeros((0, 0))
        self.residual_variance_draws = np.zeros(0)
        self.atom_probs = None

    def save(self, filename):
        """
        Save the object to a pickle file.

        Args:
          filename: The name of the file in which to save the state of this
            object.

        Effect:
          The state of this object is written to 'filename'.
        """
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """
        Load a previously fit model from a pickle file.

        Args:
          filename:  The name of the file from which to load state.

        Effects:
          The state of this object is overwritten by the saved state in
          'filename'.
        """
        with open(filename, "rb") as f:
            return pickle.load(f)

    @property
    def nclusters(self):
        """
        The number of clusters in the mixture model.
        """
        if self._model is None:
            return 0
        else:
            return self._model.nclusters

    @property
    def xdim(self):
        """
        The dimension of the dummy variable expansion of the categorical
        variables.
        """
        if self._model is None:
            return 0
        else:
            return self._model.xdim

    @property
    def ydim(self):
        """
        The number of numeric variables.
        """
        if self._numeric_colnames is None:
            return 0
        return len(self._numeric_colnames)

    @property
    def niter(self):
        """
        The number of iterations in the model.
        """
        return self.coefficient_draws.shape[0]

    def find_atoms(self, data: pd.DataFrame):
        """
        Fill a dict associating each numeric column name in data with a flat
        numpy array containing the atoms in that variable.

        Args:
          data: The data frame in which to search for atoms.

        Effect:
          self._numeric_colnames:
            If already populated then it is left unchanged.  Otherwise, it is
            filled with a list of strings giving the names of the numeric
            columns in 'data'.
          self._atoms: Filled with a dict containing the atoms in each of the
            numeric columns.  The dict is keyed by column names of the numeric
            columns.  A numeric column with no atoms is marked by an empty
            list.
        """
        if self._numeric_colnames is None:
            self._discover_column_types(data)

        for vname in self._numeric_colnames:
            variable = data[vname]
            counts = variable.value_counts().sort_values(ascending=False)
            number_observed = counts.sum()
            atom_indicator = (counts > .05 * number_observed)
            self._atoms[vname] = counts[atom_indicator].index.tolist()
            if len(self._atoms[vname]) > 3:
                self._atoms[vname] = self._atoms[vname][:3]
        print(f"Atoms: {pd.Series(self._atoms)}")

    def find_levels(self, data: pd.DataFrame):
        """
        Fill in a dict associating each categorical column name in a data frame
        with the levels of the categorical variable contained in the column.

        Effect:
          self._categorical_colnames: If already populated then it is left
            unchanged.  Otherwise it is filled with a list of strings giving
            the names fo the categorical variables in 'data'.
          self._levels: Filled with a dict containing the levels of each
            categorical column in 'data'.  The dict is keyed by variable names,
            and its elements are lists of strings.
        """
        if self._categorical_colnames is None:
            self._discover_column_types(data)

        for vname in self._categorical_colnames:
            variable = data[vname]
            levels = variable.unique()
            self._levels[vname] = levels.tolist()

    def set_level_prior(self, variable_name: str, prior_counts: np.ndarray):
        """
        Record the prior distribution on the distribution of the level
        probabilities for the specified variable.  Each cluster will use these
        prior counts as the prior distribution for this variable.

        Args:
          variable_name: A string matching the name of one of the categorical
            variables.
          prior_counts: A 1-d numpy array.  All entries must be positive, and
            the number of entries must match the number of levels for the
            specified variable.
        """
        if not np.all(prior_counts > 0):
            raise Exception("All prior counts must be positive.")
        if variable_name not in self._categorical_colnames:
            raise Exception(f"Could not find {variable_name} in categorical "
                            f"column names: {self._categorical_colnames}")

    def set_atom_prior(self, variable_name: str, prior_counts: np.ndarray):
        """
        Record the prior distribution on the distribution of atom values for the
        specified variable.  Each cluster entry uses this as its prior
        distribution.

        Args:
          variable_name: A string matching the name of one of the numeric
            variables.
          prior_counts: A 1-d numpy array.  If there are 'k' atoms then the
            array needs k+1 entries, with the last entry corresponding to the
            continuous atom.  A zero or negative entry indicates a strong prior
            assumption that the corresponding atom is never the true value.

        Effect:
          An entry is registered in self._atom_prior.  These entries are
          promoted to posterior samplers when the model is trained.
        """
        if np.sum(prior_counts > 0) == 0:
            raise Exception("At least one prior count must be positive.")
        if np.any(prior_counts < 0):
            raise Exception("Prior counts cannot be negative.")
        if variable_name not in self._atoms.keys():
            msg = "Attempt to set a prior for unrecognized variable "
            msg += f"{variable_name}."
            raise Exception
        if len(prior_counts) != len(self._atoms[variable_name]) + 1:
            raise Exception("The dimension of 'prior_counts' must exceed the"
                            " number of atoms by 1.")
        self._atom_prior[variable_name] = prior_counts.ravel()

    def train_model(self,
                    data: pd.DataFrame,
                    nclusters: int,
                    niter: int = 1000,
                    checkpoint_filename="",
                    nthreads=1):
        """
        Train the imputation model by taking 'niter' draws from the posterior
        distribution given a set of training data.

        Args:
          data: The data on which to train the model.  Any non-numeric
            variables are treated as categorical.  Missing values are expected
            to be coded as np.NaN.
          nclusters: The number of clusters to use for the joint
            distribution of the categorical data.
          niter:  The number of MCMC iterations to use during training.
          checkpoint_filename: The name of a file to use when recording the
            model's state.  The model will checkpoint its history every so
            often during training.
          nthreads: The number of threads to use during training.  This is
            experimental.

        Effects:
          self._coefficient_draws: Created and populated by MCMC draws of the
            linear regression coefficients.

        """
        self._start_time = time.time()
        if self._numeric_colnames is None or self._categorical_colnames is None:
            self._discover_column_types(data)

        # Find the set of atoms for each numeric variable.
        if self._atoms is None:
            self.find_atoms(data.loc[:, self._numeric_colnames])
        if len(self._atoms) != len(self._numeric_colnames):
            raise Exception(
                "One atom vector is needed for each numeric variable.")

        # Find the set of levels for each categorical variable.
        if self._levels is not None:
            self.find_levels(data.loc[:, self._categorical_colnames])
        if len(self._levels) != len(self._categorical_colnames):
            raise Exception(
                "Each categorical column must have its own list of levels.")

        # Convert the atoms to BOOM::Vector objects.
        atom_vector = []
        for vname in self._numeric_colnames:
            atoms = np.array(self._atoms[vname])
            atom_vector.append(boom.Vector(atoms.flatten().astype("float")))

        # The data table as a BOOM object.
        data_table = boom.to_data_table(data)

        self._model = boom.MixedDataImputer(
            nclusters, data_table, atom_vector, boom.GlobalRng.rng)

        # if nthreads > 1:
        #     self._model.setup_worker_pool(nthreads)

        print("setting prior")
        self._set_prior()

        print("Allocating space")
        self._allocate_space(niter)

        for i in range(niter):
            if i % 100 == 0:
                sep = "=-=-=-=-=-=-=-=-="
                print(f"{sep} {time.asctime()} Iteration {i} {sep}")
            self._model.sample_posterior()
            self._record_draws(i)
            if i % niter == 0 and checkpoint_filename != "":
                self.save(checkpoint_filename)
        self._end_time = time.time()

    def impute_rows(self, data, iterations):
        """
        Args:
          data:  Data frame containing the rows to impute.
          iterations:  An array-like collection of iteration numbers to use.

        Returns:
          A list of imputed data frames.

        Example:
        """
        imputed = []
        table = boom.to_data_table(data)
        for it in iterations:
            self._restore_parameters(it)
            imputed.append(
                boom.to_pandas(
                    self._model.impute_data_set(table, burn=20)))
        return imputed

    def _discover_column_types(self, data):
        """
        Populate self._numeric_colnames with a list containing the names of the
        numeric columns in 'data'.
        """
        self._numeric_colnames = []
        self._categorical_colnames = []

        for vname in data.columns:
            maybe_numeric = is_numeric_dtype(data.loc[:, vname])
            if maybe_numeric:
                counts = data.loc[:, vname].value_counts()
                if counts.shape[0] > 20:
                    self._numeric_colnames.append(vname)
            if self._numeric_colnames[-1] != vname:
                self._categorical_colnames.append(vname)

    def _format_imputation_data(self, data):
        dummies = self.encode(data)
        numerics = data.loc[:, self._numeric_colnames]
        formatted_data = []
        for i in range(data.shape[0]):
            y = boom.Vector(
                numerics.iloc[i, :].values.flatten().astype("float"))
            x = boom.Vector(dummies[i, :].flatten().astype("float"))
            formatted_data.append(boom.MvRegData(y, x))
        return formatted_data

    def _allocate_space(self, niter):
        """
        Allocate space to hold MCMC draws.

        Args:
          niter:  The desired number of draws.
        """
        self.coefficient_draws = np.empty((niter, self.xdim, self.ydim))
        self.residual_variance_draws = np.empty((niter, self.ydim, self.ydim))
        self.atom_probs = {}
        self.atom_error_probs = {}

        for vname, atom in self._atoms.items():
            k = len(atom)
            self.atom_probs[vname] = np.empty((niter, self.nclusters, k + 1))
            self.atom_error_probs[vname] = np.empty(
                (niter, self.nclusters, k + 1, k + 2))

    def _record_draws(self, i):
        """
        Record draws of the model parameters at iteration i.
        """
        self.coefficient_draws[i, :, :] = self._model.coefficients.to_numpy()
        self.residual_variance_draws[i, :, :] = (
            self._model.residual_variance.to_numpy()
        )
        for cluster in range(self.nclusters):
            for col in range(len(self._numeric_colnames)):
                name = self._numeric_colnames[col]
                self.atom_probs[name][i, cluster, :] = (
                    self._model.atom_probs(cluster, col).to_numpy()
                )

    def _restore_parameters(self, i):
        self._model.set_coefficients(boom.Matrix(
            self.coefficient_draws[i, :, :]))
        self._model.set_residual_variance(boom.Matrix(
            self.residual_variance_draws[i, :, :]))
        for cluster in range(self.nclusters):
            for col in range(len(self._numeric_colnames)):
                name = self._numeric_colnames[col]
                probs = self.atom_probs[name][i, cluster, :]
                self._model.set_atom_probs(cluster, col, boom.Vector(probs))

                error_probs = self.atom_error_probs[name][i, cluster, :, :]
                self._model.set_atom_error_probs(
                    cluster, col, boom.Matrix(error_probs))

    def _set_prior(self):
        """
        Set the prior distribution on the imputation model, and assign a
        PosteriorSampler.

        A nearly noninformative prior is chosen for the
        residual variance matrix and regression coefficients.  Each cell in in
        the mixture model is assigned an identical prior.
        """
        if self._model is None:
            raise Exception(
                "_set_prior was called before the model was created.")

        self._set_regression_prior()
        self._set_mixing_distribution_prior()
        for i, vname in enumerate(self._numeric_colnames):
            if vname not in self._atom_prior.keys():
                self._atom_prior[vname] = self._default_atom_prior(
                    self._atoms[vname])
            self._model.set_atom_prior(
                boom.Vector(self._atom_prior[vname].astype("float")),
                i)

        for i, vname in enumerate(self._categorical_colnames):
            if vname not in self._level_prior.keys():
                self._level_prior[vname] = self._default_level_prior(
                    self._levels[vname])
            self._model.set_level_prior(
                boom.Vector(self._level_prior[vname].astype("float")),
                i)

    def _set_mixing_distribution_prior(self):
        prior_counts = np.ones(self.nclusters)
        prior_counts /= len(prior_counts)
        self._model.set_mixing_weight_prior(boom.Vector(
            prior_counts.astype("float")))

    def _set_regression_prior(self):
        coefficient_prior_mean = np.zeros((self.xdim, self.ydim))
        coefficient_prior_mean[:, 0] = 0.0    # TODO replace 0.0 with ybar

        coefficient_weight = 1.0
        variance_weight = float(self.ydim + 1)
        residual_variance_guess = np.eye(self.ydim)

        self._model.set_regression_prior(
            boom.Matrix(coefficient_prior_mean.astype("float")),
            float(coefficient_weight),
            boom.SpdMatrix(residual_variance_guess.astype("float")),
            float(variance_weight))

    def _default_atom_prior(self, atoms):
        """
        Return a default prior distribution for the vector of atoms.
        """
        prior_counts = np.ones(len(atoms) + 1, dtype="float")
        prior_counts /= np.sum(prior_counts)
        return prior_counts

    def _default_level_prior(self, levels):
        """
        Return a default prior distribution for the vector of levels.
        """
        prior_counts = np.ones(len(levels), dtype="float")
        prior_counts /= np.sum(prior_counts)
        return prior_counts

    def __getstate__(self):
        """
        Return a dict that can be used to pickle a MixedDataImputer.
        """
        state = {
            "atoms": self._atoms,
            "numeric_colnames": self._numeric_colnames,
            "categorical_colnames": self._categorical_colnames,
            "atom_prior": self._atom_prior,
            "level_prior": self._level_prior,
            "coefficient_draws": self.coefficient_draws,
            "residual_variance_draws": self.residual_variance_draws,
            "atom_probs": self.atom_probs,
            "nclusters": self.nclusters,
            "empirical_distributions": self._model.empirical_distributions,
        }
        return state

    def __setstate__(self, state):
        """
        Retrieve a MixedDataImputer from a pickle.
        """
        self._atoms = state["atoms"]
        self._numeric_colnames = state["numeric_colnames"]
        self._categorical_colnames = state["categorical_colnames"]
        self._atom_prior = state["atom_prior"]
        self._dataset_encoder = state["dataset_encoder"]
        self.coefficient_draws = state["coefficient_draws"]
        self.residual_variance_draws = state["residual_variance_draws"]
        self.atom_probs = state["atom_probs"]
        self.empirical_distributions = state["empirical_distributions"]

        if (
                self._dataset_encoder is not None
                and state["nclusters"] is not None
        ):
            atom_vector = []
            for vname in self._numeric_colnames:
                atoms = np.array(self._atoms[vname])
                atom_vector.append(boom.Vector(
                    atoms.flatten().astype("float")))
            self._model = boom.MixedDataImputer(
                # TODO: this is hosed.
            )
            # state["nclusters"], atom_vector, xdim
            self._model.set_empirical_distributions(
                state["empirical_distributions"])



import scipy.stats as ss
import numpy as np


def sd(x):
    """
    Return the standard deviation of x.  If x is a matrix then return standard
    deviation of its columns.
    """
    if hasattr(x, "shape") and len(x.shape) == 2:
        return np.std(x, ddof=1, axis=0)
    else:
        return np.std(x, ddof=1)


def density(x):
    """Returns a kernel density estimate of x.  Suitable for use with
    lines_gaussian_kde in plots.py.

    """
    return ss.gaussian_kde(x)



import unittest
import numpy as np
import pandas as pd
import pickle
import json

import pdb

import matplotlib.pyplot as plt

from BayesBoom.R import delete_if_present

from BayesBoom.bsts import (
    Bsts,
    AirPassengers,
    StudentLocalLinearTrendStateModel,
    simulate_student_local_linear_trend
)


def default_colnames(n):
    return ["X" + str(i + 1) for i in range(n)]


class TestStudentLocalLinearTrend(unittest.TestCase):
    def setUp(self):
        sample_size = 100
        self._trend = simulate_student_local_linear_trend(nsteps=sample_size)
        self._y = self._trend + np.random.randn(sample_size)

    def tearDown(self):
        delete_if_present("student_llt.pkl")

    def test_local_level(self):
        model = Bsts()
        model.add_state(StudentLocalLinearTrendStateModel(self._y))
        model.train(data=self._y, niter=1000)

        fname = "student_llt.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(model, pkl)

        with open(fname, "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertEqual(model.time_dimension, m2.time_dimension)
        self.assertIsInstance(m2, Bsts)



    def test_plots(self):
        pass

    def test_predictions(self):
        pass

    def test_serialization(self):
        pass


_debug_mode = False


if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestDynamicRegression()
    # rig = TestGaussianTimeSeries()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_local_level()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import pandas as pd
import numpy as np

import BayesBoom.boom as boom
import BayesBoom.R as R


class DirichletProcessMvn:
    def __init__(self,
                 data,
                 concentration_parameter=1,
                 mean_base_measure: R.MvnGivenSigma = None,
                 variance_base_measure: R.WishartPrior = None):
        data = np.array(data, dtype="float")
        self._model = boom.DirichletProcessMvnModel(
            data.shape[1], float(concentration_parameter))

        # Assign data
        for i in range(data.shape[0]):
            self._model.add_data(R.to_boom_vector(data[i, :]))

        self._mean_base_measure = self.set_mean_base_measure(
            mean_base_measure, data)
        self._variance_base_measure = self.set_variance_base_measure(
            variance_base_measure, data)
        self._concentration_parameter = float(concentration_parameter)

        # Assign sampler.
        self._sampler = boom.DirichletProcessMvnCollapsedGibbsSampler(
            self._model,
            self._mean_base_measure.boom(),
            self._variance_base_measure.boom(),
            boom.GlobalRng.rng)

        self._model.set_method(self._sampler)

    @property
    def dim(self):
        return self._mean_base_measure.dim

    def mcmc(self, niter, ping=None, permute_state_labels=True, seed=None):
        """
        Run 'niter' iterations of MCMC.
        """
        niter = int(niter)
        if ping is None:
            ping = int(niter / 10)
            if ping < 0:
                ping = 1
        ping = int(ping)

        if seed is not None:
            boom.GlobalRng.rng.seed(seed)

        self._allocate_space(niter)
        for i in range(niter):
            R.print_timestamp(i, ping)
            self._model.sample_posterior()
            self._record_draws(i)

        self._gather_params()

        if permute_state_labels:
            self.choose_permutation()

        return None

    def set_mean_base_measure(self, mean_base_measure: R.MvnGivenSigma, data):
        """
        Args:
          mean_base_measure:
        """
        if mean_base_measure is None:
            ybar = data.mean(axis=0)
            mean_base_measure = R.MvnGivenSigma(ybar, 1)
        if not isinstance(mean_base_measure, R.MvnGivenSigma):
            raise Exception("mean_base_measure must either be None or an "
                            "instance of R.MvnGivenSigma.")
        return mean_base_measure

    def set_variance_base_measure(self,
                                  variance_base_measure: R.WishartPrior,
                                  data):
        """
        Args:
          variance_base_measure:  An instance of
        """
        if variance_base_measure is None:
            V = np.cov(data, rowvar=False, ddof=1)
            dim = V.shape[0]
            variance_base_measure = R.WishartPrior(dim + 1, V)
        if not isinstance(variance_base_measure, R.WishartPrior):
            raise Exception("variance_base_measure must either be None or an "
                            "instance of R.WishartPrior.")
        return variance_base_measure

    @property
    def log_likelihood(self):
        return np.array(self._log_likelihood)

    def choose_permutation(self):
        """
        Permute the model parameters and cluster labels to remove label
        switching.
        """
        for nclusters in self._params.keys():
            permutations = identify_permutation_from_labels(
                self._params[nclusters]["cluster_labels"])
            means = self._params[nclusters]["means"]
            variances = self._params[nclusters]["variances"]
            cluster_labels = self._params[nclusters]["cluster_labels"]
            niter = means.shape[0]
            for i in range(niter):
                perm = permutations[i, :]
                means[i, :, :] = means[i, perm, :]
                variances[i, :, :, :] = variances[i, perm, :, :]
                permuted_labels = np.arange(nclusters)[perm]
                cluster_labels[i] = permuted_labels[self._params[nclusters][
                    "cluster_labels"][i]]
            self._params[nclusters]["means"] = means
            self._params[nclusters]["variances"] = variances
            self._params[nclusters]["cluster_labels"] = cluster_labels

    def cluster_size_distribution(self, burn=None):
        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)
        burn = int(burn)
        dist = {}
        for k in self._params.keys():
            dist[k] = np.sum(np.array(self._params[k]["iterations"]) >= burn)
        dist = pd.Series(dist)
        total = np.sum(dist)
        if total > 0:
            dist = dist / total
        return dist.sort_index()

    def cluster(self, data: np.ndarray, nclusters: int, prob: bool = False,
                burn: int = None):
        """
        Return the most likely cluster (or the cluster probabilities) of one or
        more previously unseen data points.

        Args:
          data: One or more data points.  Multiple data points are rows in a 2D
            array.
          nclusters:  The number of clusters to assume for the clustering.
          prob: If True, then the return value is a numpy array giving the
            probability of cluster 1, 2, etc.  Otherwise the most likely
            cluster for each data point is returned.
          burn: The number of MCMC iterations to remove as burn-in.  The
             default (None) indicates that

        Return:
          A numpy array.
        """

        data = np.array(data)
        if len(data.shape) == 1:
            data = data.reshape((1, -1))
        nobs = data.shape[0]

        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)

        means = self._params[nclusters]["means"]
        variances = self._params[nclusters]["variances"]
        cluster_labels = self._params[nclusters]["cluster_labels"]
        if burn > 0:
            keep = self._params[nclusters]["iterations"] >= burn
            means = means[keep, :, :]
            variances = variances[keep, :, :, :]
            cluster_labels = cluster_labels[keep, :]

        niter = means.shape[0]
        if niter == 0:
            raise Exception(
                f"No clusters of size {nclusters} survived burn-in.")

        prior = np.apply_along_axis(R.table, 1, cluster_labels)
        # niter x nclusters
        total = prior.sum(axis=1).reshape((-1, 1))
        prior = prior / total

        full_cluster_prob = np.empty((niter, nobs, nclusters))

        for i in range(niter):
            logprob = np.array([np.log(prior[i, :])] * nobs)
            for k in range(nclusters):
                logprob[:, k] += R.dmvn(
                    data, means[i, k, :], variances[i, k, :, :], logscale=True)
            logprob_max = np.max(logprob, axis=1).reshape((-1, 1))
            logprob -= logprob_max
            local_prob = np.exp(logprob)
            total = local_prob.sum(axis=1).reshape((-1, 1))
            full_cluster_prob[i, :, :] = local_prob / total

        cluster_probs = full_cluster_prob.mean(axis=0)

        if prob:
            return cluster_probs
        else:
            return np.argmax(cluster_probs, axis=1)

    def _allocate_space(self, niter):
        """
        Create space for an additional 'niter' draws.
        """
        if not hasattr(self, "_params"):
            self._params = {}
        if not hasattr(self, "_log_likelihood"):
            self._log_likelihood = []

    def _ensure_params(self, number_of_clusters):
        """
        Allocate space to record a draw that has 'number_of_clusters' clusters.
        """
        if number_of_clusters not in self._params.keys():
            self._params[number_of_clusters] = {
                "means": [],
                "variances": [],
                "iterations": [],
                "cluster_labels": [],
            }

    def _record_draws(self, iteration: int):
        """
        Record the state of the current draw.

        Args:
          iteration:  The iteration number of the current draw.
        """

        number_of_clusters = self._model.number_of_clusters
        self._ensure_params(number_of_clusters)

        means = np.empty((number_of_clusters, self.dim))
        variances = np.empty((number_of_clusters, self.dim, self.dim))
        for i in range(number_of_clusters):
            means[i, :] = self._model.cluster(i).mu.to_numpy()
            variances[i, :, :] = self._model.cluster(i).Sigma.to_numpy()
        self._params[number_of_clusters]["means"].append(means)
        self._params[number_of_clusters]["variances"].append(variances)
        self._params[number_of_clusters]["iterations"].append(iteration)
        self._params[number_of_clusters]["cluster_labels"].append(
            self._model.cluster_labels)
        self._log_likelihood.append(self._model.log_likelihood)

    def _gather_params(self):
        for nclusters in self._params.keys():
            self._params[nclusters]["means"] = np.array(
                self._params[nclusters]["means"])
            self._params[nclusters]["variances"] = np.array(
                self._params[nclusters]["variances"])
            self._params[nclusters]["iterations"] = np.array(
                self._params[nclusters]["iterations"])
            self._params[nclusters]["cluster_labels"] = np.array(
                self._params[nclusters]["cluster_labels"])
        self._log_likelihood = np.array(self._log_likelihood)


def _compute_mean_cluster_probs(cluster_indicators, permutations):
    """
    Args:
      cluster_indicators: niter x nobs x nclusters array of 0/1 cluster
        indicators.  Element (i, j, k) is 1 if in MCMC iteration i observation
        j was assigned to cluster k.  The element is 0 otherwise.
      permutations: niter x nclusters array of permutations.  Each row is an
        arrangement of the nubmers 0..nclusters-1.
    """
    niter, nobs, nclusters = cluster_indicators.shape
    if permutations.shape[0] != niter:
        raise Exception("permutations has the wrong leading dimension.")
    permuted_indicators = cluster_indicators.copy()
    for i in range(niter):
        permuted_indicators[i, :, :] = cluster_indicators[
            i, :, permutations[i, :]].T

    prior = np.full((1, nobs, nclusters), 1.0 / nclusters)
    permuted_indicators = np.concatenate((permuted_indicators, prior))
    probs = permuted_indicators.mean(axis=0)
    return probs


def _solve_linear_assignment_problem(cluster_indicators: np.ndarray,
                                     mean_probs: np.ndarray):
    """
    Let nobs denote the number of observations, and nclusters the number of
    clusters.

    cluster_indicators: A [nobs x nclusters] array of 0/1 dummy variables
      indicating the cluster assigned to each observation.
    mean_probs:  The current probability

    Returns:
      permutation, total_cost
    """
    nclusters = mean_probs.shape[1]
    cost = np.empty((nclusters, nclusters))
    log_probs = np.log(mean_probs)

    for i in range(nclusters):
        for j in range(nclusters):
            # cost[i, j] is -1 times the contribution to multinomial log
            # likelihood of cluster labels == j under probability column i.
            cost[i, j] = -1 * cluster_indicators[:, i].dot(log_probs[:, j])

    lap = boom.LinearAssignment(boom.Matrix(cost))
    min_cost = lap.solve()
    permutation = lap.row_solution
    return permutation, min_cost


def identify_permutation_from_labels(cluster_labels):
    """
    Args:
      cluster_labels: A [niter x nobs] numpy array.  Element [i, j] is the
        integer cluster label for observation j in Monte Carlo iteration i.

    Returns:
      permutations:  A [niter x nclusters] numpy array of integers.
        cluster_labels[i, permutations[i, :]]
    """
    cluster_labels = np.array(cluster_labels)
    niter, nobs = cluster_labels.shape
    nclusters = int(cluster_labels.max()) + 1
    levels = np.arange(nclusters)
    encoder = R.OneHotEncoder("blah", levels=levels, baseline_level=None)

    # cluster_indicators are the dummy variable representation of
    # cluster_labels.
    cluster_indicators = np.empty((niter, nobs, encoder.dim))
    for i in range(niter):
        cluster_indicators[i, :, :] = encoder.encode(cluster_labels[i, :])

    permutations = np.array([np.arange(nclusters)] * niter).reshape(
        (niter, nclusters))

    total_cost = np.inf
    cost_reduction = np.inf
    while cost_reduction > 1e-5:
        mean_cluster_probs = _compute_mean_cluster_probs(
            cluster_indicators, permutations)

        old_total_cost = total_cost
        total_cost = 0
        for draw in range(niter):
            permutation, min_cost = _solve_linear_assignment_problem(
                cluster_indicators[draw, :, :], mean_cluster_probs)
            total_cost += min_cost
            permutations[draw, :] = permutation
        cost_reduction = old_total_cost - total_cost

    return permutations



import unittest
import numpy as np
import pandas as pd

from BayesBoom.R import cbind


class TestCbind(unittest.TestCase):

    def setUp(self):
        np.random.seed(8675309)

    def test_cbind_numpy(self):
        x = np.random.randn(3)
        ans = cbind(1, x)
        self.assertEqual(ans.shape, (3, 2))
        self.assertTrue(np.allclose(ans[:, 1], x))
        self.assertTrue(np.allclose(ans[:, 0], np.full(3, 1.0)))

        X = np.random.randn(3, 4)
        ans = cbind(x, X, 1.0)
        self.assertEqual(ans.shape, (3, 6))
        self.assertTrue(np.allclose(ans[:, 0], x))
        self.assertTrue(np.allclose(ans[:, 1], X[:, 0]))
        self.assertTrue(np.allclose(ans[:, 2], X[:, 1]))
        self.assertTrue(np.allclose(ans[:, 5], np.ones(3)))

        ans = cbind(x, x)
        self.assertEqual(ans.shape, (3, 2))
        self.assertTrue(np.allclose(ans[:, 0], x))
        self.assertTrue(np.allclose(ans[:, 1], x))

    def test_cbind_pandas(self):
        x = np.random.randn(3)
        y = pd.Series(np.random.randn(3))
        z = 2.0
        w = np.random.randn(3, 4)
        d = pd.DataFrame(np.random.randn(3, 2), columns=["Fred", "Barney"])
        foo = cbind(x, y)
        self.assertTrue(isinstance(foo, pd.DataFrame))
        self.assertTrue(foo.shape, (3, 2))
        self.assertTrue(np.allclose(foo.iloc[:, 0], x))
        self.assertTrue(np.allclose(foo.iloc[:, 1], y))

        bar = cbind(x, y, z)
        self.assertEqual(bar.shape, (3, 3))
        self.assertTrue(np.allclose(bar.iloc[:, 2], np.full(3, 2.0)))

        baz = cbind(x, y, z, w)
        self.assertEqual(baz.shape, (3, 7))
        self.assertTrue(np.allclose(baz.iloc[:, 3], w[:, 0]))

        qux = cbind(x, y, z, w, d)
        self.assertEqual(qux.shape, (3, 9))
        self.assertTrue(np.allclose(qux.iloc[:, 7], d.iloc[:, 0]))


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestCbind()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_cbind_numpy()
    rig.test_cbind_pandas()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import unittest
import numpy as np
import pickle


from BayesBoom.bsts import (
    Bsts,
    AirPassengers,
    ArStateModel,
    AutoArStateModel,
    LocalLinearTrendStateModel,
    SeasonalStateModel,
)

from BayesBoom.R import delete_if_present


class TestAr(unittest.TestCase):
    def setUp(self):
        self._y = np.log(AirPassengers)

    def tearDown(self):
        delete_if_present("ar.pkl")

    def test_basic_ar(self):
        model = Bsts()
        model.add_state(LocalLinearTrendStateModel(self._y))
        model.add_state(SeasonalStateModel(self._y, nseasons=12))
        model.add_state(ArStateModel(self._y, lags=1))
        model.train(data=self._y, niter=1000)

        fname = "ar.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(model, pkl)

        with open(fname, "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertEqual(model.time_dimension, m2.time_dimension)
        self.assertIsInstance(m2, Bsts)

        pred = model.predict(12)

    def test_auto_ar(self):
        model = Bsts()
        model.add_state(LocalLinearTrendStateModel(self._y))
        model.add_state(SeasonalStateModel(self._y, nseasons=12))
        model.add_state(AutoArStateModel(self._y, lags=1))
        model.train(data=self._y, niter=1000)

        fname = "ar.pkl"
        with open(fname, "wb") as pkl:
            pickle.dump(model, pkl)

        with open(fname, "rb") as pkl:
            m2 = pickle.load(pkl)

        self.assertEqual(model.time_dimension, m2.time_dimension)
        self.assertIsInstance(m2, Bsts)

    def test_predictions(self):
        pass


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestAr()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_basic_ar()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import matplotlib.pyplot as plt
from abc import ABC, abstractmethod


_active_graphics_devices = {}
_current_graphics_device = None
_largest_graphics_device_number = 0


class GraphicsDevice(ABC):
    """
    Manages a plt.figure and a set of axes.
    """

    def __init__(self):
        """
        Creating a new graphics device updates the global set of devices
        """
        global _largest_graphics_device_number
        global _current_graphics_device
        global _active_graphics_devices
        self._number = _largest_graphics_device_number + 1
        _largest_graphics_device_number = self._number
        _active_graphics_devices[self._number] = self
        _current_graphics_device = self

        self._figure, self._axes = plt.subplots(1, 1)
        self._axes_cursor = None
        self._nrow = 1
        self._ncol = 1

    def __del__(self):
        plt.close(self._figure)
        global _active_graphics_devices
        del _active_graphics_devices[self._number]

    @property
    def next_axes(self):
        """
        The next set of axes on which a plot is to be drawn.  If the last set of
        available axes has been exhausted, the stored Figure is refreshed with
        new axes of the same dimension.
        """
        if hasattr(self._axes, "shape"):
            shape = self._axes.shape
            if min(shape) == 1:
                self._increment_1d_cursor()
            else:
                self._increment_2d_cursor()
        else:
            self._axes = self._figure.subplots(1, 1)
        return self.current_axes

    @property
    def current_axes(self):
        """
        The current set of axes.
        """
        if self._nrow > 1 or self._ncol > 1:
            return self._axes[self._axes_cursor]
        else:
            return self._axes

    @abstractmethod
    def draw_current_axes(self):
        """
        A hook to be called after the graphics device is updated.
        """

    def _create_subplots(self, nrow, ncol):
        """
        Set the graphics device to use multiple rows and/or columns of plots.
        """
        self._axes = self._figure.subplots(nrow, ncol)
        self._nrow = nrow
        self._ncol = ncol
        if min(nrow, ncol) > 1:
            self._axes_cursor = (0, 0)
        elif max(nrow, ncol) > 1:
            self._axes_cursor = 0
        else:
            self._axes_cursor = None

    def _increment_1d_cursor(self):
        """
        Increment the self._axes cursor when self._axes is a 1-d array.
        """
        dim = max(self._nrow, self._ncol)
        self._axes_cursor += 1
        if self._axes_cursor >= dim:
            self._axes_cursor = 0
            # We need nrow and ncol here because a 1-d axes might be a row, or
            # a column.
            self._axes = self._figure.subplots(self._nrow, self._ncol)

    def _increment_2d_cursor(self):
        """
        Increment self._axes_cursor when self._axes is two-dimensional.
        """
        i, j = self._axes_cursor
        j += 1
        if j >= self._axes.shape[1]:
            j = 0
            i += 1
            if i >= self._axes.shape[0]:
                self._axes = self._figure.subplots(
                    self._axes.shape[0], self._axes.shape[1])
                i, j = 0, 0
        self._axes_cursor = (i, j)


def dev_new():
    return InteractiveGraphicsDevice()


def dev_set(device_number: int):
    """
    Set the current graphics device to the device with the given number.

    """
    global _active_graphics_devices
    global _current_graphics_device
    device = _active_graphics_devices.get(device_number, None)
    if device is None:
        raise Exception(f"Graphics device {device_number} does note exist.")
    else:
        _current_graphics_device = device


class InteractiveGraphicsDevice(GraphicsDevice):
    """
    A private stack manages the set of active graphics devices.
    """

    def __init__(self):
        super().__init__()
        plt.show(block=False)
        plt.pause(.001)

    def draw_current_axes(self):
        """
        """
        plt.pause(.001)


class PdfGraphicsDevice(GraphicsDevice):
    """
    Destructor generates a PDF file that gets generated when the graphics
    device is deleted.
    """

    def __init__(self, filename, width=5, height=5):
        self._filename = filename

    def __del__(self):
        """
        Create the pdf file upon deletion.
        """
        if self._figure is not None:
            self._figure.save(self._filename)

    def draw_current_axes(self):
        """
        """
        pass


def get_current_graphics_device():
    """
    Returns the current graphics device, if one exists.  Otherwise create and
    return an interactive graphics device.
    """
    global _current_graphics_device
    if _current_graphics_device is not None:
        return _current_graphics_device
    else:
        return InteractiveGraphicsDevice()



import numpy as np
from abc import ABC, abstractmethod
import copy

"""
Wrapper classes to encapsulate and expand models and prior distributions
from the Boom library.
"""


class DoubleModel(ABC):
    """
    A base class that marks its children as being able to produce a
    boom.DoubleModel, which is simply a model that implements a 'logp' method
    measuring a real valued random variable.
    """

    @abstractmethod
    def boom(self):
        """
        Return a boom.DoubleModel with parameters set from this object.
        """

    @property
    @abstractmethod
    def mean(self):
        """
        The mean of the distribution.
        """


class SdPrior(DoubleModel):
    """A prior distribution for a standard deviation 'sigma'.  This prior assumes
    that 1/sigma**2 ~ Gamma(a, b), where a = df/2 and b = ss/2.  Here 'df' is
    the 'sample_size' and ss is the "sum of squares" equal to the sample size
    times 'sigma_guess'**2.

    This prior allows an upper limit on the support of sigma, which is infinite
    by default.

    """

    def __init__(self, sigma_guess, sample_size=.01, initial_value=None,
                 fixed=False, upper_limit=np.inf):
        """
        Create an SdPrior.

        Args:
          sigma_guess:  Guess at the value of the standard deviation.
          sample_size: Number of observations worth of information with which
            to weight the guess.
          initial_value: The initial value to be used in an MCMC chain.  This
            is not always respected.  The default value is sigma_guess.
          fixed: Flag indicating whether the parameter should be held fixed in
            an MCMC algorithm.  This is mainly for debugging and is not always
            respected.
          upper_limit: Upper limit on the value of 'sigma'.
        """
        self.sigma_guess = float(sigma_guess)
        self.sample_size = float(sample_size)
        if initial_value is None:
            initial_value = sigma_guess
        self.initial_value = float(initial_value)
        self.fixed = bool(fixed)
        self.upper_limit = float(upper_limit)

    @property
    def sum_of_squares(self):
        return self.sigma_guess**2 * self.sample_size

    def create_chisq_model(self):
        return self.boom()

    def boom(self):
        """
        Return the boom.ChisqModel corresponding to the input parameters.
        """
        import BayesBoom.boom as boom
        return boom.ChisqModel(self.sample_size, self.sigma_guess)

    @property
    def mean(self):
        """
        The mean of the distribution on the precision scale.
        """
        return self.sample_size / self.sigma_guess**2

    def __repr__(self):
        ans = f"SdPrior with sigma_guess = {self.sigma_guess}, "
        ans += f"sample_size = {self.sample_size}, "
        ans += f"upper_limit = {self.upper_limit}"
        return ans

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, payload):
        self.__dict__ = payload


class NormalPrior(DoubleModel):
    """
    A scalar normal prior distribution.
    """
    def __init__(self,
                 mu: float = 0.0,
                 sigma: float = 1.0,
                 initial_value: float = None):
        self.mu = float(mu)
        self.sigma = float(sigma)
        if initial_value is None:
            self.initial_value = mu
        else:
            self.initial_value = float(initial_value)

    @property
    def mean(self):
        return self.mu

    @property
    def sd(self):
        return self.sigma

    @property
    def variance(self):
        return self.sigma ** 2

    def boom(self):
        """
        Return the boom.GaussianModel corresponding to the object's parameters.
        """
        import BayesBoom.boom as boom
        return boom.GaussianModel(self.mu, self.sigma)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, payload):
        self.__dict__ = payload


class Ar1CoefficientPrior(DoubleModel):
    """
    Contains the information needed to create a prior distribution on an AR1
    coefficient.
    """
    def __init__(self,
                 mu: float = 0.0,
                 sigma: float = 1.0,
                 force_stationary: bool = True,
                 force_positive: bool = False,
                 initial_value: float = None):
        """
        Args:
          mu: The prior mean of the coefficient.
          sigma:  The prior standard deviation of the coefficient.
          force_stationary: If True then the prior support for the AR1
            coefficient will be truncated to (-1, 1).
          force_positive: If True then the prior for the AR1 coefficient will
            be truncated to positive values.
          initial_value: A suggestion about where to start an MCMC sampling
            run.  The default is to use mu.
        """
        self.mu = mu
        self.sigma = sigma
        self.force_stationary = force_stationary
        self.force_positive = force_positive
        self.initial_value = initial_value
        if initial_value is None:
            self.initial_value = mu

    def boom(self):
        """
        Return the boom.GaussianModel corresponding to this object's
        parameters.
        """
        import BayesBoom.boom as boom
        return boom.GaussianModel(self.mu, self.sigma)

    @property
    def mean(self):
        return self.mu

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, payload):
        self.__dict__ = payload


class MvnPrior:
    """
    Encodes a multivariate normal distribution.
    """
    def __init__(self, mu, Sigma):
        if len(mu.shape) != 1:
            raise Exception("mu must be a vector.")
        if len(Sigma.shape) != 2:
            raise Exception("Sigma must be a matrix.")
        if Sigma.shape[0] != Sigma.shape[1]:
            raise Exception("Sigma must be symmetric")
        if Sigma.shape[0] != len(mu):
            raise Exception("mu and Sigma must be the same dimension.")
        self._mu = mu
        self._Sigma = Sigma

    @property
    def dim(self):
        return len(self._mu)

    @property
    def mu(self):
        return self._mu

    @property
    def mean(self):
        return self.mu

    @property
    def Sigma(self):
        return self._Sigma

    @property
    def variance(self):
        return self.Sigma

    def boom(self):
        """
        Return the boom.MvnModel corresponding to this object's parameters.
        """
        import BayesBoom.boom as boom
        return boom.MvnModel(boom.Vector(self._mu),
                             boom.SpdMatrix(self._Sigma))


class MvnGivenSigma:
    """
    Encodes a conditional multivariate normal distribution given an external
    variance matrix Sigma.  This model describes y ~ Mvn(mu, Sigma / kappa).
    """
    def __init__(self, mu: np.ndarray, sample_size: float):
        self._mu = np.array(mu, dtype="float").ravel()
        self._sample_size = float(sample_size)

    @property
    def dim(self):
        return len(self._mu)

    def boom(self):
        import BayesBoom.boom as boom
        return boom.MvnGivenSigma(self._mu, self._sample_size)


class UniformPrior(DoubleModel):
    """
    Univariate uniform distribution.
    """
    def __init__(self, lo, hi):
        if hi < lo:
            lo, hi = hi, lo
        self._lo = lo
        self._hi = hi

    @property
    def mean(self):
        return .5 * (self._lo + self._hi)

    def boom(self):
        """
        Return the boom.UniformModel corresponding to this object's parameters.
        """
        import BayesBoom.boom as boom
        return boom.UniformModel(self._lo, self._hi)


class BetaPrior(DoubleModel):
    """
    A distribution, typically used as the prior over a scalar probability.
    """
    def __init__(self, a=1.0, b=1.0):
        self._a = float(a)
        self._b = float(b)

    @property
    def mean(self):
        return self._a / (self._a + self._b)

    def boom(self):
        import BayesBoom.boom as boom
        return boom.BetaModel(self._a, self._b)


class DirichletPrior:
    """
    A Dirichlet prior distribution over discrete probability distributions.
    """

    def __init__(self, counts):
        counts = np.array(counts)
        if not np.all(counts > 0):
            raise Exception("All elements of 'counts' must be positive.")
        self._counts = counts

    def boom(self):
        import BayesBoom.boom as boom
        return boom.DirichletModel(boom.Vector(self._counts))


class WishartPrior:
    def __init__(self, df: float, variance_estimate: np.ndarray):
        """
        Args:
          df: The prior sample size.  For the distribution to be proper df must
            be larger than the number of rows in 'variance_estimate'.
          variance_estimate: A symmetric positive definite matrix defining the
            center of the distribution.

        Let X_i ~ Mvn(0, V).  Then the Wishart(nu, V) distribution describes
        the sum of 'nu' draws X_i * X_i'.  If the draws are placed as rows in a
        matrix X then X'X ~ Wishart(nu, V).  The mean of this distribution is
        nu * V.

        The Wishart distribution is the conjugate prior for the precision
        parameter (inverse variance) of the multivariate normal distribution.
        """
        sumsq = df * variance_estimate
        if len(sumsq.shape) != 2:
            raise Exception("sumsq must be a matrix")

        if sumsq.shape[0] != sumsq.shape[1]:
            raise Exception("sumsq must be square")

        sym_sumsq = (sumsq + sumsq.T) * .5
        sumabs = np.sum(np.abs(sumsq - sym_sumsq))
        relative = np.sum(np.abs(sumsq))
        if sumabs / relative > 1e-8:
            raise Exception("sumsq must be symmetric")

        if df <= sumsq.shape[0]:
            raise Exception(
                "df must be largern than nrow(sumsq) for the prior to be "
                "proper.")

        self._df = df
        self._sumsq = sumsq

    @property
    def variance_estimate(self):
        return self._sumsq / self._df

    @property
    def df(self):
        return self._df

    def boom(self):
        import BayesBoom.boom as boom
        return boom.WishartModel(self.df, self.variance_estimate)


class GaussianSuf:
    """
    Sufficient statistics for a scalar normal model.
    """

    def __init__(self, data=None):
        """
        Args:
          data: If None (the default) then an empty GaussianSuf is created.
            Otherwise create a new GaussianSuf summarizing 'data'.
        """
        self._sum = 0
        self._sumsq = 0
        self._n = 0
        if data is not None:
            self.update(data)

    def update(self, incremental_data):
        """
        Add summaries of the incremental data to the data already summarized.

        Args:
          data:  A 1-d numpy array, or equivalent.

        Effects:
          The sufficient statistics in the object are updated to describe data.
        """
        y = np.array(incremental_data)
        self._sum += np.nansum(y)
        self._sumsq += np.nansum(y * y)
        self._n += np.sum(~np.isnan(incremental_data))

    def combine(self, other):
        """
        Add the sufficient statistics from 'other' to 'self'.  This operation is
        done inplace.  The 'other' object is unaffected.
        """
        self._n += other._n
        self._sum += other._sum
        self._sumsq += other._sumsq

    def __iadd__(self, other):
        """
        Implements operator +=.  Other can either be a GaussianSuf or raw data.
        """
        if isinstance(other, GaussianSuf):
            self.combine(other)
        else:
            self.update(other)
        return self

    def __add__(self, other):
        """
        Implements operator+.  Other can either be a GaussianSuf or raw data.
        """
        ans = copy.copy(self)
        ans += other
        return ans

    @property
    def sample_size(self):
        return self._n

    @property
    def sum(self):
        return self._sum

    @property
    def mean(self):
        if self.sample_size > 0:
            return self._sum / self.sample_size
        else:
            return 0.0

    @property
    def sumsq(self):
        return self._sumsq

    def centered_sumsq(self, center=None):
        if self.sample_size <= 0:
            return 0
        if center is None:
            center = self.mean
        n = self.sample_size
        return self.sumsq - 2 * center * self.sum + n * center ** 2

    @property
    def sample_sd(self):
        return np.sqrt(self.sample_variance)

    @property
    def sample_variance(self):
        n = self.sample_size
        if n < 2:
            return 0
        return self.centered_sumsq() / (n - 1)



import BayesBoom.boom as boom
import numpy as np
import BayesBoom.R as R
from .state_models import StateModel


class GeneralSeasonalLLT(StateModel):
    """
    A seasonal state model that builds a local trend into the seasonal values.
    """

    def __init__(self,
                 y,
                 nseasons: int,
                 initial_state_prior=None,
                 level_precision_priors=None,
                 slope_precision_priors=None,
                 sdy: float = None):
        """
        Args:
          y: The time series to be modeled.  This can be "None" if 'sdy' is
            supplied.
          nseasons: The number of seasons in a cycle.
          initial_state_prior: An R.NormalPrior object describing the initial
            distribution of the state at time 0.  If None then a default prior
            will be assumed.
          level_precision_priors: A list of R.SdPrior objects describing the
            prior distribution on the innovation standard deviations for the
            level portion of the model.  There is one such prior for each
            season in the cycle.
          slope_precision_priors: A list of R.SdPrior objects describing the
            prior distribution on the innovation standard deviations for the
            slope portion of the model.  There is one such prior for each
            season in the cycle.
          sdy: The standard deviation of the time series to be modeled.  This
            is not needed if 'y' is supplied, or if all the prior distributions
            are explicity supplied.
        """

        self._nseasons = int(nseasons)
        if nseasons <= 1:
            raise Exception("Seasonal models require at least 2 seasons.")

        if initial_state_prior is None:
            if sdy is None:
                sdy = self._compute_sdy(y, "initial_state_prior")
            initial_state_prior = self._default_initial_state_prior(sdy)

        if isinstance(initial_state_prior, R.NormalPrior):
            dim = 2 * self._nseasons
            mu = np.zeros(dim)
            sigma = initial_state_prior.sd
            Sigma = np.diag(np.ones(dim) * sigma ** 2)
            self._initial_state_prior = R.MvnPrior(mu, Sigma)
        else:
            self._initial_state_prior = initial_state_prior

        if not isinstance(self._initial_state_prior, R.MvnPrior):
            raise Exception("intial_state_prior must be an R.MvnPrior")

        if level_precision_priors is None:
            if sdy is None:
                sdy = self._compute_sdy(y, "level_precision_priors")
            self._level_precision_priors = [R.SdPrior(
                sdy / 100, .1, upper_limit=sdy) for i in range(self.nseasons)]
        else:
            self._level_precision_priors = level_precision_priors
        msg = "level_precision_priors must be a list of R.SdPrior objects"
        if not isinstance(self._level_precision_priors, list):
            raise Exception(msg)
        for x in self._level_precision_priors:
            if not isinstance(x, R.SdPrior):
                raise Exception(msg)

        if slope_precision_priors is None:
            if sdy is None:
                sdy = self._compute_sdy(y, "slope_precision_priors")
            self._slope_precision_priors = [R.SdPrior(
                sdy / 100, .1, upper_limit=sdy) for i in range(self.nseasons)]
        else:
            self._slope_precision_priors = slope_precision_priors
        msg = "slope_precision_priors must be a list of R.SdPrior objects"
        if not isinstance(self._slope_precision_priors, list):
            raise Exception(msg)
        for x in self._slope_precision_priors:
            if not isinstance(x, R.SdPrior):
                raise Exception(msg)

        self._build_model()
        self._state_contribution = None

    @property
    def label(self):
        return f"SeasonalTrend({self._nseasons})"

    @property
    def nseasons(self):
        """
        The number of seasons in a full cycle.
        """
        return self._nseasons

    @property
    def state_dimension(self):
        """
        The number of elements in the full state vector.
        """
        return self._nseasons * 2

    @property
    def state_contribution(self):
        """
        The posterior distribution of the contribution of this state component
        to the mean of y.
        """
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self._sigma_level_draws = np.empty((niter, self.nseasons))
        self._sigma_slope_draws = np.empty((niter, self.nseasons))
        self._state_contribution = np.zeros((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self._sigma_level_draws[iteration, :] = (
            self._state_model.sigma_level.to_numpy())
        self._sigma_slope_draws[iteration, :] = (
            self._state_model.sigma_slope.to_numpy())
        time_index = state_matrix.shape[1]
        nseasons = self.nseasons
        row_index = np.array(list(range(self.nseasons)) * int(
            np.ceil(time_index / nseasons)))[:time_index] + self.state_index
        col_index = range(state_matrix.shape[1])
        self._state_contribution[iteration, :] = state_matrix[
            row_index, col_index]

    def restore_state(self, iteration):
        self._state_model.set_sigma_level(self._sigma_level_draws[iteration, :])
        self._state_model.set_sigma_slope(self._sigma_slope_draws[iteration, :])

    def plot_state_contribution(
            self, fig, gridspec, time, burn=None, ylim=None, **kwargs):
        return self.plot_state_contribution_default(
            fig=fig, gridspec=gridspec, time=time, burn=burn,
            ylim=ylim, **kwargs)

    def _build_model(self):
        self._state_model = boom.GeneralSeasonalLLT(self._nseasons)
        self._state_model.set_initial_state_mean(
            boom.Vector(self._initial_state_prior.mu))
        self._state_model.set_initial_state_variance(
            boom.SpdMatrix(self._initial_state_prior.variance))

        level_precision_priors = [
            x.boom() for x in self._level_precision_priors]
        slope_precision_priors = [
            x.boom() for x in self._slope_precision_priors]

        sampler = boom.GeneralSeasonalLLTIndependenceSampler(
            self._state_model,
            level_precision_priors,
            slope_precision_priors,
            boom.GlobalRng.rng)
        self._state_model.set_method(sampler)

    def _default_initial_state_prior(self, sdy):
        dim = 2 * self._nseasons
        mean = np.zeros(dim)
        variance = np.diag(np.ones(dim) * sdy)
        return R.MvnPrior(mean, variance)

    def _compute_sdy(self, y, which_prior: str):
        if y is None:
            raise Exception(
                f"One of y, sdy, or {which_prior} must be specified.")
        return np.nanstd(y, ddof=1)

    def __repr__(self):
        return f"A GeneralSeasonalLLT model with {self.nseasons} seasons."

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_model()



import unittest
import BayesBoom.R as R
import numpy as np
import pandas as pd
import BayesBoom.boom as boom

import pdb

def char_range(c1, c2):
    """Generates the characters from `c1` to `c2`, inclusive."""
    for c in range(ord(c1), ord(c2) + 1):
        yield chr(c)


# Some categorical data to use for testing.
LETTERS = list(char_range("a", "z"))

# fmt: off
us_states = ["AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DC", "DE", "FL", "GA",
             "HI", "ID", "IL", "IN", "IA", "KS", "KY", "LA", "ME", "MD", "MA",
             "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH", "NJ", "NM", "NY",
             "NC", "ND", "OH", "OK", "OR", "PA", "RI", "SC", "SD", "TN", "TX",
             "UT", "VT", "VA", "WA", "WV", "WI", "WY"]
# fmt: on


def random_words(sample_size, word_length=10):
    np.random.seed(8675309)
    return ["".join(np.random.choice(LETTERS, size=word_length))
            for i in range(sample_size)]


class DataTableTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(8675309)
        boom.GlobalRng.rng.seed(8675309)

        sample_size = 100
        numerics = np.random.randn(sample_size, 3)
        colors = np.random.choice(
            ["red", "blue", "green"], size=sample_size, replace=True)
        shapes = np.random.choice(
            ["circle", "square", "triangle"], size=sample_size, replace=True)
        self._data = pd.DataFrame({
                "X1": numerics[:, 0],
                "colors": colors,
                "X2": numerics[:, 1],
                "shapes": shapes,
                "X3": numerics[:, 2]
            })

    def create_base_dataset(
            self, sample_size, num_numeric, num_cat, num_levels):
        xdim = 1 + num_cat * (num_levels - 1)
        cats = {}
        levels = {}
        encoders = []
        for i in range(num_cat):
            # local_levels are the levels this variable can assume.
            local_levels = random_words(num_levels)
            vname = "cat" + str(i+1)
            values = np.random.choice(local_levels, sample_size)
            cats[vname] = values
            levels[vname] = local_levels
            encoders.append(boom.EffectsEncoder(i, local_levels))

        ydim = num_numeric
        self._beta = np.random.randn(xdim, ydim)

        Rho = boom.random_correlation_matrix(ydim).to_numpy()
        S = np.diag(R.rgamma(ydim, 1, 1))
        self._Sigma = S @ Rho @ S
        Sigma_root = np.linalg.cholesky(self._Sigma)

        errors = (Sigma_root @ np.random.randn(ydim, sample_size)).T
        encoder = boom.DatasetEncoder(encoders)
        xcat = encoder.encode_dataset(R.to_data_table(
            pd.DataFrame(cats)))
        yhat = xcat.to_numpy() @ self._beta
        numerics = yhat + errors
        self._data = pd.DataFrame(
            numerics,
            columns=["X" + str(i + 1) for i in range(ydim)]
        )
        for vname, column in cats.items():
            self._data[vname] = column
        self._ydim = ydim
        self._xdim = xdim
        self._ncat = num_cat

    def modify_dataset(self, nclusters):
        pass

    def test_data_table(self):
        table = R.to_data_table(self._data)
        self.assertEqual(table.nrow, self._data.shape[0])
        self.assertEqual(table.ncol, self._data.shape[1])

        frame = R.to_data_frame(table)
        for i in range(5):
            self.assertTrue(np.all(self._data.iloc[:, i] == frame.iloc[:, i]))

    def test_autoclean_construction(self):
        ac = R.AutoClean()
        ac.train_model(self._data, nclusters=3, niter=10)

    def test_mcmc(self):
        nlevels = 5
        self.create_base_dataset(
            sample_size=1000,
            num_numeric=3,
            num_cat=4,
            num_levels=nlevels)
        ac = R.AutoClean()
        niter = 100
        nclusters = 7
        ac.train_model(self._data, nclusters=nclusters, niter=niter)

        beta = self._beta
        xdim = beta.shape[0]
        ydim = beta.shape[1]
        self.assertEqual(ac.coefficients.shape, (niter, xdim, ydim))
        self.assertEqual(ac.residual_variance.shape, (niter, ydim, ydim))

        self.assertEqual(len(ac.atom_probs), ydim)
        self.assertEqual(ac.atom_probs["X1"].shape,
                         (niter, nclusters, 1))
        self.assertTrue(
            np.allclose(ac.atom_probs["X1"].sum(axis=2),
                        np.ones((niter, nclusters)))
        )

        self.assertEqual(len(ac.atom_error_probs), ydim)
        self.assertEqual(ac.atom_error_probs["X1"].shape,
                         (niter, nclusters, 1, 2))
        self.assertTrue(np.allclose(
            ac.atom_error_probs["X1"].sum(axis=3),
            np.ones((niter, nclusters, 1))))

        self.assertEqual(len(ac.level_probs), self._ncat)
        self.assertEqual(ac.level_probs["cat1"].shape,
                         (niter, nclusters, nlevels))
        self.assertTrue(np.allclose(
            ac.level_probs["cat1"].sum(axis=2),
            np.ones((niter, nclusters))))

        self.assertEqual(len(ac.level_observation_probs), self._ncat)
        self.assertEqual(ac.level_observation_probs["cat1"].shape,
                         (niter, nclusters, nlevels, nlevels+1))
        self.assertTrue(np.allclose(
            ac.level_observation_probs["cat1"].sum(axis=3),
            np.ones((niter, nclusters, nlevels))))

        to_impute = self._data.iloc[[3, 9, 24], :]
        iterations = [50, 60, 70, 75, 90]

        imputed = ac.impute_rows(to_impute, iterations)
        self.assertEqual(len(imputed), len(iterations))
        self.assertEqual(imputed[0].shape, to_impute.shape)



_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = DataTableTest()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_mcmc()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import numpy as np
import pandas as pd
import patsy
import BayesBoom.boom as boom
import BayesBoom.R as R
import scipy.sparse

from .priors import RegressionSpikeSlabPrior


def sparsify(glm_coefs):
    # Convert a boom.GlmCoefs objects to a 1-row sparse matrix.
    inc = glm_coefs.inc.included_positions
    zeros = np.zeros(len(inc))
    return scipy.sparse.csr_matrix(
        (glm_coefs.included_coefficients.to_numpy(),
         (zeros, inc)),
        shape=(1, glm_coefs.inc.nvars_possible))


def set_glm_coefs(glm_coefs: boom.GlmCoefs,
                  sparse_coefs: scipy.sparse.lil_matrix):
    """
    Set the internal values of a boom.GlmCoefs to match a 1-row sparse
    matrix.

    Args:
      glm_coefs:  The boom object to be set.
      sparse_coefs: The value to set.

    """
    nonzero_positions = sparse_coefs.nonzero()[1]
    nonzero_values = np.array(sparse_coefs.data[0])
    glm_coefs.set_sparse_coefficients(nonzero_values, nonzero_positions)


def dot(data_frame, omit=[]):
    """
    Build a formula string by "summing" all entries except those on an 'omit
    list'.  This would typically include the name of the variable on the left
    hand side of the equation.

    This function is named for the 'dot' operator in R, where a formula given
    as 'y ~ .' means "regress y on all other variables".  The R dot operator
    can also be used to specify interactions, as in y ~ .^2.  To allow for
    similar specifications, the return value of this function is wrapped in
    paraentheses "()".

    Args:
      data_frame: The data frame from which to build the equation.  A list or
        array of column names is also acceptable.

      omit: A list of names (strings) to omit.  As a convenience, if a single
        variable is to be omitted a single string can be passed instead of a
        list containing that single string.

    Returns:
      A string containing the list of names in data_frame, separated by '+'.
      The return value begins with '(' and ends with ')' so that y~dot(data,
      omit=["y"])**2 can be used to specify all 2-way interactions.

    Examples:
      formula = "y ~ " + dot(my_data_frame, "y")
      # Returns "y ~ (X1 + X2 + X3 + extraneous_user_id)"

      formula = "y ~ " + dot(my_data_frame, ["y", "extraneous_user_id"])
      # Returns "y ~ (X1 + X2 + X3)"

      formula = f"y ~ {dot(my_data_frame, omit=["y", "extraneous_user_id"])}**2
      # Returns "y ~ (X1 + X2 + X3)**2"

    """

    # Allow 'omit' to be a string, for the common case where there is just one
    # name to omit.
    if isinstance(omit, str):
        omit = [omit]

    vnames = [x for x in data_frame.columns if x not in omit]
    ans = "(" + "+".join(x for x in vnames) + ")"
    return ans


class lm_spike:
    """
    Fit a linear model with a spike and slab prior using MCMC.

    Typical use:

    from boom.spikeslab import lm_spike
    from sklearn.model_selection import train_test_split

    data = pd.read_csv("mydata.csv")
    train_data, pred_data = train_test_split(data, test_size=100)

    model = lm_spike('y ~ .', niter=1000, data=train_data)
    pred = model.predict(pred_data)

    model.plot()
    model.plot("coefficients")
    model.plot("inc")
    model.plot("resid")

    model.summary()
    model.residuals()

    pred.plot()

    """

    def __init__(self,
                 formula: str,
                 niter: int,
                 data: pd.DataFrame,
                 prior: RegressionSpikeSlabPrior = None,
                 ping: int = None,
                 seed: int = None,
                 **kwargs):
        """Create and a model object and run a specified number of MCMC iterations.

        Args:
          formula: A model formula that can be interpreted by the 'patsy'
            module to produce a model matrix from 'data'.
          niter: The desired number of MCMC iterations.
          data: A pd.DataFrame containing the data with which to train the
            model.
          prior: A SpikeSlabPrior object providing the prior distribution over
            the inclusion indicators, the coefficients, and the residual
            variance parameter.
          ping: The frequency (in iterations) with which to print status
            updates.  If ping is None then niter/10 will be assumed.
          seed: The seed for the C++ random number generator, or None.
          **kwargs: Extra argumnts will be passed to SpikeSlabPrior.

        Returns:
          An lm_spike object.

        """

        response, predictors = patsy.dmatrices(formula, data, eval_env=1)
        self._x_design_info = predictors.design_info
        # xdim = predictors.shape[1]
        # sample_size = predictors.shape[0]
        niter = int(niter)
        if niter <= 0:
            raise Exception("niter should be a positive integer.")

        if ping is None:
            ping = int(niter / 10)
        ping = int(ping)

        if seed is not None:
            boom.GlobalRng.rng.seed(int(seed))

        X = boom.Matrix(predictors)
        y = boom.Vector(response)
        nvars = X.ncol

        self._model = boom.RegressionModel(X, y, False)
        if prior is None:
            prior = RegressionSpikeSlabPrior(x=X, y=y, **kwargs)

        sampler = boom.BregVsSampler(
            self._model,
            prior.slab(self._model.Sigsq_prm),
            prior.residual_precision,
            prior.spike)
        self._model.set_method(sampler)
        # A lil matrix is a "linked list" matrix.  This is an efficient method
        # for constructing matrices.  It should be converted to a different
        # matrix type before doing anything with it.
        self._coefficient_draws = scipy.sparse.lil_matrix((niter, nvars))
        self._residual_sd = np.zeros(niter)
        self._log_likelihood = np.zeros(niter)

        for i in range(niter):
            self._model.sample_posterior()
            self._residual_sd[i] = self._model.sigma
            beta = self._model.coef
            self._coefficient_draws[i, :] = sparsify(beta)
            self._log_likelihood[i] = self._model.log_likelihood()

        # Convert the coefficient draws to sparse column format.  Predictions
        # vs this format should take the form X @ beta, not beta @ X.
        self._coefficient_draws = self._coefficient_draws.tocsc()

        self._fitted_values = self.predict(predictors).mean(axis=0)
        self._residuals = y.to_numpy() - self._fitted_values

    @property
    def xdim(self):
        return self._model.xdim

    @property
    def log_likelihood(self):
        return self._log_likelihood

    @property
    def residuals(self):
        return self._residuals

    @property
    def fitted_values(self):
        return self._fitted_values

    @property
    def xnames(self):
        # A list of strings containing the column names of the predictors.
        return self._x_design_info.column_names

    def inclusion_probs(self, burn=None):
        """
        Args:
          burn:  The number of MCMC iterations to discard as burn-in.

        Returns:
          A pd.Series containing the marginal probability that each coefficient
          is nonzero.  The series is indexed by the set of variable names.

        """
        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)
        probs = compute_inclusion_probabilities(
            self._coefficient_draws[burn:, ])
        return pd.Series(probs, index=self.xnames)

    def coefficient_positive_probability(self, burn=None):
        """
        Args:
          burn:  The number of MCMC iterations to discard as burn-in.

        Returns:
          A pd.Series containing the marginal probability that each coefficient
          is positive.

        """
        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)
        probs = coefficient_positive_probability(
            self._coefficient_draws[burn:, :])
        return pd.Series(probs, index=self.xnames)

    def plot(self, what=None, **kwargs):
        """Plot an aspect of the model.

        Args:
          what: The type of plot desired.  Acceptable choices are
            "inclusion", "coefficients", "residual", and "predicted".

          kwargs: Extra arguments are passed to the specific plot function
            being called.

        """

        plot_types = ["inclusion", "coefficients", "residual", "predicted"]
        if what is None:
            what = plot_types[0]
        what = R.unique_match(what, plot_types)
        if what == "coefficients":
            return self.plot_coefficients(**kwargs)
        elif what == "inclusion":
            return self.plot_inclusion(**kwargs)
        elif what == "residual":
            return self.plot_residual(**kwargs)
        elif what == "predicted":
            return self.plot_predicted(**kwargs)
        else:
            raise Exception(f"Unknown plot type {what}.")

    def plot_inclusion(self, burn=None, inclusion_threshold=0,
                       unit_scale=True, number_of_variables=None,
                       ax=None, **kwargs):
        """
        A barplot showing the marginal inclusion probability of each variable.

        Args:
          burn:
        """
        inc = self.inclusion_probs(burn=burn)
        pos = self.coefficient_positive_probability(burn=burn)
        colors = np.array([str(x) for x in pos])
        index = np.argsort(inc.values)[::-1]

        if number_of_variables is None:
            number_of_variables = np.sum(inc >= inclusion_threshold)
        inc = inc[index[:number_of_variables]]
        pos = pos[index[:number_of_variables]]
        colors = colors[index[:number_of_variables]]
        ans = R.barplot(inc,
                        ax=ax,
                        color=colors[::-1],
                        linewidth=.25,
                        edgecolor="black",
                        xlab="Marginal Inclusion Probability",
                        ylab="Variable",
                        **kwargs)
        return ans

    def plot_coefficients(self, **kwargs):
        """A boxplot showing the values of the coefficients.
        """

    def plot_residual(self, hexbin_threshold=1e+5,
                      xlab="fitted", ylab="residual"):
        """A plot of the residuals vs the predicted values.

        """
        fig, ax = R.plot(self.fitted_values, self.residuals,
                         hexbin_threshold=hexbin_threshold,
                         xlab=xlab, ylab=ylab)
        if len(self.residuals) > hexbin_threshold:
            abs_resid = np.abs(self.residuals)
            n = len(abs_resid) - 100
            top_resids = np.argpartition(abs_resid, n)[n:]
            ax.scatter(self.fitted_values[top_resids],
                       self.residuals[top_resids],
                       s=5,
                       color="yellow")
        ax.axhline(color="black", linewidth=.5)
        return fig, ax

    def plot_predicted(self, xlab="fitted", ylab="actual"):
        """A plot of predicted values vs actual values.

        """
        fig, ax = R.plot(self.fitted_values,
                         self.residuals + self.fitted_values,
                         xlab=xlab,
                         ylab=ylab)
        return fig, ax

    def predict(self, newdata, burn=None, seed=None):
        """
        Return an LmSpikePrediciton object.
        """
        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)
        if seed is not None:
            boom.GlobalRng.rng.seed(int(seed))
        if isinstance(newdata, np.ndarray) and len(newdata.shape) == 1:
            newdata = newdata.reshape(1, -1)
        if isinstance(newdata, np.ndarray) and newdata.shape[1] == self.xdim:
            predictors = newdata
        else:
            predictors = patsy.build_design_matrices(
                [self._x_design_info],
                data=newdata)[0]
        return self._coefficient_draws[burn:, :] @ predictors.T

    def summary(self, burn=None):
        """Return a summary of model fit, including something like R^2, and residual
        sd, along with a table of coefficients, standard deviations, and
        marginal inclusion probabilities.

        """
        return lm_spike_summary(self)


class lm_spike_summary:
    """
    Summarizes the fit of an lm_spike model.
    """

    def __repr__(self):
        return """A spike and slab model summary!
        Put R2, residual_sd, and top 10 coefficients here.
        """


def compute_inclusion_probabilities(coefficients):
    """
    Args:
      coefficients:  A (scipy) sparse matrix of regression coefficients.
        Rows represent MCMC draws.  Columns represent variables.

    Returns:
      A np.array of inclusion probabilities.
    """
    nvars = coefficients.shape[1]
    return np.array(
        [
            np.mean(coefficients[:, i] != 0) for i in range(nvars)
        ]
    )


def coefficient_positive_probability(coefficients):
    nvars = coefficients.shape[1]
    return np.array(
        [
            np.mean(coefficients[:, i] > 0) for i in range(nvars)
        ]
    )


def plot_inclusion_probs(coefficients, burn, xnames, inclusion_threshold=0,
                         unit_scale=True, number_of_variables=None, ax=None,
                         **kwargs):
    """
    """
    coef = coefficients[burn:, :]
    inc = compute_inclusion_probabilities(coef)
    pos = coefficient_positive_probability(coef)
    colors = np.array([str(x) for x in pos])
    index = np.argsort(inc.values)[::-1]

    if number_of_variables is None:
        number_of_variables = np.sum(inc >= inclusion_threshold)
    inc = inc[index[:number_of_variables]]
    pos = pos[index[:number_of_variables]]
    colors = colors[index[:number_of_variables]]
    ans = R.barplot(inc,
                    ax=ax,
                    color=colors[::-1],
                    linewidth=.25,
                    edgecolor="black",
                    xlab="Marginal Inclusion Probability",
                    ylab="Variable",
                    **kwargs)
    return ans


def plot_model_size(coefficients, burn, ax=None, **kwargs):
    ndraws = coefficients.shape[0]
    size = np.array([
        np.sum(coefficients[i, :] != 0)
        for i in range(burn, ndraws)
    ])
    return R.hist(size, ax=ax, **kwargs)


class RegressionSlabPrior:
    """
    A multivariate normal distribution intended to be the prior in a multiple
    regression problem.  The prior is

    beta ~ N(b, V)

    where b = (ybar, 0, 0, 0, ....)
    and V^{-1} = kappa * [(1 - alpha) * xtx + alpha * diag(xtx)] / n

    The mean parameter shrinks the intercept towards the sample mean, and all
    other coefficients towards zero.  In the literature it is more standard to
    shrink all coefficients towards zero, but in practice this can inflate
    estimates of the residual standard deviation.

    The prior precision is defined in terms of xtx: the cross product matrix
    from the regression problem.  We average xtx with its diagonal (with weight
    alpha on the diagonal) to ensure that the overall matrix is full rank.  xtx
    is the information matrix for the regression coefficients in a standard
    regression problem, so dividing by 'n' (the sample size) turns the whole
    thing into the "average information from a single observation."
    Multiplying by 'kappa' means that the information content of the prior is
    equivalent to 'kappa' prior observations.
    """

    def __init__(self, xtx, sample_mean, data_sample_size,
                 prior_sample_size=1.0, diagonal_shrinkage=0.05):
        """
        Args:
          Please see the class comments, above.
          xtx:  The cross product matrix from the regression.
          sample_mean: The mean of the response variable in the regression
            problem ('ybar' above).
          data_sample_size: The number of observations in the regression ('n'
            above).
          prior_sample_size: The number of observations of prior weight to
            assign the prior.  ('kappa' above).
          diagonal_shrinkage: The weight to assign the diagonal of xtx in the
            full rank adjustment.  ('alpha' above).
        """
        self._xtx = xtx
        self._sample_mean = sample_mean
        self._data_sample_size = data_sample_size
        self._prior_sample_size = prior_sample_size
        self._diagonal_shrinkage = diagonal_shrinkage

    def set_xtx(self, xtx: np.ndarray):
        self._xtx = xtx

    def boom(self, sigsq_param: boom.UnivParams):
        xtx = self._xtx if self._xtx is not None else np.ones((1, 1))
        return boom.RegressionSlabPrior(
            boom.SpdMatrix(xtx),
            sigsq_param,
            self._sample_mean,
            self._data_sample_size,
            self._prior_sample_size,
            self._diagonal_shrinkage)


class BigAssSpikeSlab:
    """
    A regression trained using a spike and slab regression.  This class differs
    from lm_spike in that we don't expect the full data set to be stored in
    memory.

    The BigAssSpikeSlab requires the data to be streamed through the model
    twice.  The first stream is an initial screen used to identify candidate
    variables.

    Because the model is based on streaming data, encoding data frames into
    numeric matrices of predictors needs to happen outside the model object.
    """

    def __init__(self,
                 xdim: int,
                 subordinate_model_max_dim: int = 500,
                 force_intercept: bool = True,
                 spike=None,
                 slab: RegressionSlabPrior = None,
                 residual_sd_prior: R.SdPrior = None,
                 expected_model_size=1.0,
                 expected_Rsqure=0.5,
                 prior_sample_size=1.0,
                 seed: int = None,
                 **kwargs):
        """
        Args:
          xdim:  The dimension of the large sparse vector of predictors.
          subordinate_model_max_dim: Each subordinate model will manage at most
            this many predictors.
          force_intercept: Should each subordinate model be required to have an
            intercept term?

        """
        self._xdim = xdim
        self._subordinate_model_max_dim = subordinate_model_max_dim
        self._force_intercept = force_intercept

        self._model = boom.BigRegressionModel(
            xdim, subordinate_model_max_dim, force_intercept)

        if slab is None:
            self._slab = RegressionSlabPrior(None, np.nan, -1)
        else:
            self._slab = slab
        if not isinstance(self._slab, RegressionSlabPrior):
            raise Exception("slab must be a RegressionSlabPrior")

        if spike is None:
            self._spike = np.full(xdim, expected_model_size / xdim)
        else:
            self._spike = np.array(spike)

        self._residual_sd_prior = residual_sd_prior
        if residual_sd_prior is not None:
            if not isinstance(residual_sd_prior, R.SdPrior):
                raise Exception("residual_sd_prior must be an R.SdPrior")

        self._response_suf = boom.GaussianSuf()

        self._sampler = None
        self._expected_Rsqure = expected_Rsqure
        self._prior_sample_size = prior_sample_size

    @property
    def xdim(self):
        return self._xdim

    def stream_data_for_initial_screen(self, x: np.ndarray, y: np.ndarray):
        """
        Pass the data to the underlying C++ model object for the purpose of
        running an initial screen.

        Arg:
          x: Matrix of predictor variables.  If an intercept term is desired it
            should be present in the first column.  Any dummy variables and
            basis expansions (e.g. splines) should already be included.
          y:  The response vector.
        """
        for i, yi in enumerate(y):
            data_point = boom.RegressionData(yi, boom.Vector(x[i, :]))
            self._model.stream_data_for_initial_screen(data_point)
            self._response_suf.increment(boom.Vector(y))

    def initial_screen(self,
                       niter=1000,
                       threshold: float = .05,
                       max_candidates_per_model: int = 1000,
                       use_threads: bool = True):
        """
        Run an initial screen to identify candidate variables.  Under the covers
        there are multiple models running independent spike and slab samplers
        to identify sets of candidate variables.

        Args:
          niter: The number of MCMC iterations that should be used in the
            initial screen by each subordinate model.
          threshold: A probability.  Predictors with marginal inclusion
            probabilities above this threshold in their subordinate models will
            be promoted as candidates for inclusion in the second round.
          max_candidates_per_model: If the number of predictor variable
            candidates in a subordinate model exceeds this number then the
            candidate list will be truncated to this number.
          use_threads: If True then the underlying C++ code will run the screen
            in parallel using C++11 threads.  Set this to True unless you've
            got a really good reason not to.
        """
        self._ensure_priors()
        self._sampler.initial_screen(
            niter, threshold, use_threads)

    def _ensure_priors(self):
        if self._residual_sd_prior is None:
            sample_var = self._response_suf.sample_var
            residual_var = (1 - self._expected_Rsqure) * sample_var
            self._residual_sd_prior = R.SdPrior(
                np.sqrt(residual_var), self._prior_sample_size)

        if self._sampler is None:
            self._sampler = boom.BigAssSpikeSlabSampler(
                self._model,
                boom.VariableSelectionPrior(self._spike),
                self._slab.boom(self._model.Sigsq_prm),
                self._residual_sd_prior.boom())
            self._model.set_method(self._sampler)

    def stream_data_for_restricted_model(self, x: np.ndarray, y: np.ndarray):
        """
        After the initial_screen has been run, the data will need to be
        streamed a second time.  The arguments here are identical to
        'stream_data_for_initial_screen'.
        """
        for i, yi in enumerate(y):
            data_point = boom.RegressionData(yi, boom.Vector(x[i, :]))
            self._model.stream_data_for_restricted_model(data_point)

    def train(self, niter):
        """
        This function should only be called after 'initial_screen' and
        'stream_data_for_restricted_model' have completed.

        Run the MCMC algorithm on the potential candidates identified by the
        initial screen.
        """
        self._allocate_space(niter)
        for i in range(niter):
            self._sampler.draw()
            self._residual_sd[i] = self._model.sigma
            beta = self._model.coef
            self._coefficient_draws[i, :] = sparsify(beta)
            self._log_likelihood[i] = self._model.log_likelihood()

    def _allocate_space(self, niter):
        # A lil matrix is a "linked list" matrix.  This is an efficient method
        # for constructing matrices.  It should be converted to a different
        # matrix type before doing anything with it.
        self._coefficient_draws = scipy.sparse.lil_matrix((niter, self.xdim))
        self._residual_sd = np.zeros(niter)
        self._log_likelihood = np.zeros(niter)



import BayesBoom.boom as boom
import numpy as np
import patsy
import BayesBoom.R as R
from numbers import Number
import matplotlib.pyplot as plt
# import BayesBoom.spikeslab as spikeslab
# import scipy.sparse


class SparseDynamicRegressionModel:
    """
    A DynamicRegressionModel is a time series regression model where the
    coefficients obey a classic state space model.  Note that the number of
    observations at each time point might differ.  The model is implemented as
    a multivariate state space model.  Through data augmentation one can extend
    this model to most GLM's.

    Define the set of responses at time t as Y'_t = [y_1t, y_2t, ... y_n_tt],
    where Y_t = X[t] * beta[t] + error[t], with temporally IID error term
    error[t] ~ N(0, Diagonal(sigma^2)).

    Each coefficient beta[j, t] is zero with probability determined by a Markov
    chain Pr(beta[j, t] = s | beta[j, t-1] = r) = q[r, s], for r, s in {0, 1}.

    The conditional distribution of beta[j, t] given beta[j, t-1], and given
    that both are nonzero, is normal with mean b_jt = T_ij b_jt-1, and variance
    tau^2.

    Expected Use:

    data = get_data_from_somewhere()
    model = SparseDynamicRegressionModel(
        y ~ dot(data, ["y", "timestamp_column"]),
        data=data,
        timestamps="timestamp_column",
        niter=1000,
    )

    """

    def __init__(self,
                 formula,
                 data,
                 timestamps,
                 niter: int = 0,
                 residual_precision_prior=None,
                 coefficient_innovation_priors=None,
                 prior_inclusion_probabilities=None,
                 expected_inclusion_duration=None,
                 transition_probability_prior_sample_size=None,
                 ping=None,
                 seed=None):
        """
        Args:
          formula: A string of the form "response ~ v1 + v2" where "y", "v1"
            and "v2" are named columns in 'data'.  See the 'dot' function in
            the BayesBoom.spikeslab package for specifying complex formulas
            simply.
          data: A pd.DataFrame or equivalent containing the data used in the
            model.
          timestamps: Either a vector of type datetime64[ns] (Is this overly
            restrictive?) containing

          innovation_precision_priors:  Prior distribution on the residual
            precision parameter 1/sigsq.  This must be an object inheriting
            from boom.GammaModelBase.  The typcial choice is a boom.ChisqModel.


        """
        from pandas import unique

        # self._timestamps indicates the timestamp corresponding each element
        # in the input data.  If there are many observations per time point
        # then self._timestamps will contain many repeat values.
        if isinstance(timestamps, str):
            self._timestamps = data[timestamps]
        else:
            self._timestamps = timestamps

        # self._unique_timestamps is the sorted set of timestamps covered by
        # the training data.
        self._unique_timestamps = unique(self._timestamps)
        self._unique_timestamps.sort()

        self._formula = formula
        self._data = data
        self._data_by_timestamp = None

        self._residual_precision_prior = residual_precision_prior
        self._coefficient_innovation_priors = coefficient_innovation_priors
        self._prior_inclusion_probabilities = prior_inclusion_probabilities
        self._expected_inclusion_duration = expected_inclusion_duration
        tmp = transition_probability_prior_sample_size
        self._transition_probability_prior_sample_size = tmp

        self._model = None

        if niter > 0:
            self.train(niter=niter, ping=ping, seed=seed)

    @property
    def xdim(self):
        """
        The number of potential predictor variables.
        """
        if self._model:
            return self._model.xdim
        elif self._data is not None and self._formula is not None:
            if "~" in self._formula:
                pred_formula = self._formula.split("~")[1]
            else:
                pred_formula = self._formula
            predictors = patsy.dmatrix(pred_formula, self._data.iloc[:2, :])
            return predictors.shape[1]
        else:
            return 0

    @property
    def time_dimension(self):
        """
        The number of time points in the model's data.
        """
        if self._model:
            return self._model.time_dimension
        return 0

    def train(self, niter: int, ping: int = None, seed: int = None):
        self._create_model()
        self._set_data(self._formula, self._data, self._timestamps)
        self._validate_priors()
        self._create_posterior_sampler(
            self._residual_precision_prior,
            self._coefficient_innovation_priors,
            self._prior_inclusion_probabilities,
            self._expected_inclusion_duration,
            self._transition_probability_prior_sample_size)

        if seed is not None:
            np.random.seed(seed)
            boom.GlobalRng.rng.seed(seed)
        self._allocate_space(niter)
        for i in range(niter):
            self._model.sample_posterior()
            self._record_state(i)

    def suggest_burn(self):
        return R.suggest_burn(-1 * self._residual_sd_draws)

    def plot(self, what="coefficients", **kwargs):

        plot_types = [
            "coefficients", "size", "residual_sd"
        ]
        what = R.unique_match(what.lower(), plot_types)
        if what == "coefficients":
            return self.plot_coefficients(**kwargs)
        elif what == "residual_sd":
            return self.plot_residual_sd(**kwargs)
        elif what == "size":
            return self.plot_model_size(**kwargs)
        else:
            raise Exception(
                f"Supplied plot_type {what} is not in the set "
                f"of supported {plot_types}."
            )

    def plot_residual_sd(self,
                         burn: int = None,
                         type: str = "density",
                         ax=None,
                         **kwargs):
        """
        Args:
          burn: The number of MCMC iterations to discard as burn-in.  "None"
            indicates that an estimated default number should be used.
          type: The type of plot.  "density" shows a kernel density estimate of
            the residual SD draws.  "ts" shows a time series plot of the draws.
          ax: A plt.Axes object on which to draw the plot.  If None new Figure
            and Axes objects are created and drawn on function exit.
          kwargs:  Further keyword arguments are ignored.

        Effects:
          A plot is added to the relevant Axes object.

        Returns:
          The Axes object on which the plot is drawn.
        """
        plot_types = ["density", "ts"]
        type = R.unique_match(type, plot_types)

        if burn is None:
            burn = self.suggest_burn()

        if burn < 0:
            burn = 0
        sd = self._residual_sd_draws[burn:]

        show_plot = False
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            show_plot = True

        if type == "density":
            density = R.Density(sd)
            density.plot(ax=ax, xlab="Residual SD", ylab="Density")
        elif type == "ts":
            iteration = np.arange(len(self._residual_sd_draws))
            if burn > 0:
                iteration = iteration[burn:]
            ax.plot(iteration, sd)
            ax.set_xlabel("Iteration")
            ax.set_ylabel("Residual SD")

        if show_plot:
            fig.show()
        return ax

    def plot_size(self, ax=None, burn: int = None, **kwargs):
        fig = None
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        size = np.sum(self._beta_draws != 0, axis=1)
        R.plot_dynamic_distribution(
            size,
            timestamps=self._unique_timestamps,
            ax=ax,
            xlab="Time",
            ylab="Number Included Predictors",
        )

        if fig is not None:
            fig.show()
        return ax

    def plot_coefficients(self, subset=None, fig=None, ylim=None,
                          same_scale: bool = True,
                          burn: int = None,
                          **kwargs):
        """
        Plot some or all of the model coefficients on the same figure.

        Args:
          subset: Identifies the subset of coefficients to be plotted. 'None'
            means plot all coefficients.  Otherwise 'subset' must be an object
            that can be used to index a numpy array, such as a vector of
            integers or booleans.
          fig: The plt.Figure object on which to draw the plots.  If None a new
            Figure will be created and drawn upon function exit.
          ylim: A pair of numbers giving the lower and upper limits of the Y
            axes.  If specified then all plots will be drawn with the same
            values.
          same_scale: This is ignored if ylim is specified.  Otherwise, True
            indicates that the Y axes should all be drawn on the same scale.
            False means each coefficient is plotted on its own scale.
          kw_args:  Additional keyword arguments are ignored.
          burn: The number of MCMC iteration to discard as burn-in.
        Returns:
          The Figure object on which the plots are drawn.
        """
        if subset is None:
            coef = self._beta_draws
        else:
            coef = self._beta_draws[:, subset, :]

        if burn is None:
            burn = self.suggest_burn()

        if burn > 0:
            coef = coef[burn:, :, :]

        call_show = fig is None
        if fig is None:
            fig = plt.figure()

        nr, nc = R.plot_grid_shape(coef.shape[1])
        sharey = same_scale or (ylim is not None)
        ax = fig.subplots(nr, nc, sharex=True, sharey=sharey)
        index = 0

        if same_scale and (ylim is None):
            ylim = R.data_range(self._beta_draws)

        for col in range(nc):
            for row in range(nr):
                if index < coef.shape[1]:
                    self.plot_single_coefficient(
                        coef[:, index, :],
                        ax=ax[row][col],
                        ylim=ylim)
                index += 1
        if call_show:
            fig.set_tight_layout(True)
            fig.show()
        return fig

    def plot_single_coefficient(self, beta, ylim=None, ax=None,
                                highlight_median="green"):
        """
        Plot the dynamic distribution of a single model coefficient.

        Args:
          beta: The coefficient to be plotted.  A matrix.  Rows are Monte Carlo
            draws, and columns are time points.
          ylim: A pair of numbers giving the lower and upper limits of the Y
            axis.  If 'None' then 'ylim' will be inferred from the range of the
            data.
          ax: A plt.Axes object on which to draw.  If None then a new
            plt.Figure and Axes will be created and drawn on function exit.
          highlight_median: The name of a color used to draw the meadian of the
            curves at each time point.  The empty string signals not to add the
            extra highlighting.

        Returns:
          The axes object containing the plot.
        """
        fig = None
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        R.plot_dynamic_distribution(beta,
                                    timestamps=self._unique_timestamps,
                                    ax=ax,
                                    ylim=ylim,
                                    highlight_median=highlight_median)
        if fig is not None:
            fig.show()
        return ax

    def _create_model(self):
        """
        Populate self._model with the Boom model object.
        """
        self._model = boom.DynamicRegressionModel(self.xdim)

    def _set_data(self, formula, data, timestamps):
        """
        Partitiion the DataFrame 'data' into chunks defined by 'timestamps',
        pass it through the 'formula', and convert the output to the
        expected BayesBoom objects.

        Args:
          formula: A string defining a formula as interpreted by the 'patsy'
            library.
          data: A data frame containing the variables in 'formula', and
            maybe other variables as well.  Extraneous variables will be
            ignored.
          timestamps: A vector-like object containing objects that can be
            ordered.  E.g. a vector of dates, or integers.  Each element of
            'timestamps' corresponds to a row of 'data', and determines
            which the time point to which that row belongs.

        Effects:
          * Data are added to the BOOM model.
          * self._response_suf is created and populated with response data.
          * self._predictor_suf is created and each element is populated with
              data from the corresponding predictor variable.
        """

        unique_time_points = sorted(set(timestamps))
        self._response_suf = R.GaussianSuf()
        xdim = self.xdim
        self._predictor_suf = [R.GaussianSuf()] * xdim
        for time_stamp in unique_time_points:
            subset = timestamps == time_stamp
            response, predictors = patsy.dmatrices(
                formula, data.loc[subset, :], eval_env=1)
            data_point = boom.RegressionDataTimePoint(
                boom.Matrix(predictors),
                boom.Vector(response))
            self._response_suf += response
            for i in range(xdim):
                self._predictor_suf[i] += predictors[:, i]

            self._model.add_data(data_point)

    def _validate_priors(self):
        self._validate_residual_precision_prior()
        self._validate_coefficient_innovation_priors()
        self._validate_inclusion_prior()

    def _validate_residual_precision_prior(self):
        """
        Ensure the self._residual_precision_prior is the correct class.

        Preconditions:
            self._set_data must have been run.
        """
        if isinstance(self._residual_precision_prior, R.SdPrior):
            return

        # Assume an expected R^2 of 50%.
        target_variance = self._response_suf.sample_variance / 2.0
        self._residual_precision_prior = R.SdPrior(
            sigma_guess=np.sqrt(target_variance),
            sample_size=1.0)

    def _validate_coefficient_innovation_priors(self):
        """
        Ensure that self._coefficient_innovation_priors are a list of SdPriors.
        """
        if (isinstance(self._coefficient_innovation_priors, list) and np.all(
                [isinstance(x, R.SdPrior) for x in
                 self._coefficient_innovation_priors])):
            return

        if isinstance(self._coefficient_innovation_priors, R.SdPrior):
            self._coefficient_innovation_priors = [
                self._coefficient_innovation_priors] * self.xdim
            return

        if self._coefficient_innovation_priors is not None:
            raise Exception("coefficient_innovation_priors must either be an "
                            "R.SdPrior or a list of such priors.")

        sdy = self._response_suf.sample_sd
        self._coefficient_innovation_priors = [
            R.SdPrior(.01 * sdy / self._predictor_suf[i].sample_sd, 1)
            for i in range(self.xdim)
        ]

    def _validate_inclusion_prior(self):
        xdim = self.xdim
        if self._expected_inclusion_duration is None:
            self._expected_inclusion_duration = [50] * xdim
        elif isinstance(self._expected_inclusion_duration, Number):
            self._expected_inclusion_duration = [
                self._expected_inclusion_duration] * xdim
        if len(self._expected_inclusion_duration) != xdim:
            raise Exception(f"Expected {xdim} numbers in expected"
                            "_inclusion_duration.")
        elif np.any([x <= 0 for x in self._expected_inclusion_duration]):
            raise Exception("All entries in expected_inclusion_duration"
                            " must be positive")
        self._expected_inclusion_duration = np.array(
            self._expected_inclusion_duration, dtype=float)

        if self._prior_inclusion_probabilities is None:
            self._prior_inclusion_probabilities = [1.0 / xdim] * xdim
        elif isinstance(self._prior_inclusion_probabilities, Number):
            self._prior_inclusion_probabilities = [
                self._prior_inclusion_probabilities] * xdim
        if len(self._prior_inclusion_probabilities) != xdim:
            raise Exception(f"Expected {xdim} elements in prior_inclusion_"
                            "probabilities.")
        elif np.any([x < 0 for x in self._prior_inclusion_probabilities]):
            raise Exception("All prior_inclusion_probabilities must be "
                            "non-negative.")
        elif np.any([x > 1 for x in self._prior_inclusion_probabilities]):
            raise Exception("All prior_inclusion_probabilities must be less"
                            "than 1.")
        self._prior_inclusion_probabilities = np.array(
            self._prior_inclusion_probabilities, dtype=float)

        if self._transition_probability_prior_sample_size is None:
            self._transition_probability_prior_sample_size = [1.0] * xdim
        elif isinstance(self._transition_probability_prior_sample_size,
                        Number):
            self._transition_probability_prior_sample_size = [
                self._transition_probability_prior_sample_size] * xdim

        if len(self._transition_probability_prior_sample_size) != xdim:
            raise Exception(f"Expected {xdim} elements in transition_"
                            "probability_prior_sample_size.")
        elif np.any([x <= 0 for x in
                     self._transition_probability_prior_sample_size]):
            raise Exception("All elements of transition_probability_prior_"
                            "sample_size must be positive.")
        self._transition_probability_prior_sample_size = np.array(
            self._transition_probability_prior_sample_size, dtype=float)

    def _create_posterior_sampler(
            self,
            residual_precision_prior,
            coefficient_innovation_priors,
            prior_inclusion_probabilities,
            expected_inclusion_duration,
            transition_probability_prior_sample_size):
        sampler = boom.DynamicRegressionDirectGibbsSampler(
            self._model,
            residual_precision_prior.sigma_guess,
            residual_precision_prior.sample_size,
            boom.Vector(np.array(
                [x.sigma_guess for x in coefficient_innovation_priors])),
            boom.Vector(np.array(
                [x.sample_size for x in coefficient_innovation_priors])),
            boom.Vector(prior_inclusion_probabilities),
            boom.Vector(expected_inclusion_duration),
            boom.Vector(transition_probability_prior_sample_size))
        self._model.set_method(sampler)

    def _allocate_space(self, niter):
        """
        Create space to store 'niter' MCMC draws.
        """
        self._beta_draws = np.zeros((niter, self.xdim, self.time_dimension))
        self._residual_sd_draws = np.zeros((niter))
        self._innovation_sd_draws = np.zeros((niter, self.xdim))
        self._transition_probabilities = np.zeros((niter, self.xdim, 2, 2))

    def _record_state(self, i):
        # TODO: all_coefficients
        self._beta_draws[i, :, :] = self._model.all_coefficients.to_numpy()
        self._residual_sd_draws[i] = self._model.residual_sd
        self._innovation_sd_draws[i, :] = (
            self._model.unscaled_innovation_sds
            * self._model.residual_sd).to_numpy()
        for pred in range(self.xdim):
            self._transition_probabilities[i, pred, :, :] = (
                self._model.transition_probabilities(pred).to_numpy()
            )



import numpy as np
import pandas as pd
import patsy
import BayesBoom.boom as boom
import BayesBoom.R as R
import scipy.sparse

from .priors import MultinomialLogitSpikeSlabPrior
from .spikeslab import sparsify


class mlogit_spike:
    """
Suppose the data looks like:
country         age     sex     married size    type
American	34	Male	Married	Large	Family
Japanese	36	Male	Single	Small	Sporty
Japanese	23	Male	Married	Small	Family
American	29	Male	Single	Large	Family
American	39	Male	Married	Medium	Family
Japanese	34	Male	Single	Medium	Family

    """

    def __init__(self,
                 response,
                 subject_formula: str,
                 choice_formula: dict,
                 niter: int,
                 data: pd.DataFrame,
                 levels: list = None,
                 prior: MultinomialLogitSpikeSlabPrior = None,
                 ping: int = None,
                 seed: int = None,
                 **kwargs):
        """
        Create and a model object and run a specified number of MCMC iterations.

        Args:
          response: Either a string naming a column in the data frame, or an
            object convertible to a 1-d numpy array of dtype "object".
          subject_formula: A model formula that can be interpreted by 'patsy'
            to produce a model matrix from 'data'.
          choice_formula: A dict of strings, keyed by levels of the response
            variable.  Dictionary entries are strings that can be fed to
            'patsy' to construct predictor matrices from 'data'.  See the
            method build_choice_formula for help constructing the formula
            semi-programmatically.
          niter: The desired number of MCMC iterations.
          data: A pd.DataFrame containing the data with which to train the
            model.
          prior: A SpikeSlabPrior object providing the prior distribution over
            the inclusion indicators, the coefficients, and the residual
            variance parameter.
          ping: The frequency (in iterations) with which to print status
            updates.  If ping is None then niter/10 will be assumed.
          seed: The seed for the C++ random number generator, or None.
          **kwargs: Extra argumnts will be passed to SpikeSlabPrior.

        Returns:
          An lm_spike object.
        """

        if not levels:
            if isinstance(choice_formula, dict) and len(choice_formula) > 0:
                levels = list(choice_formula.keys())
            else:
                levels = list(pd.unique(response))
        self._levels = levels

        subject_predictors = patsy.dmatrix(subject_formula, data, eval_env=1)
        self._subject_x_design_info = subject_predictors.design_info

        choice_predictors = {
            x: patsy.dmatrix(choice_formula[x], data, eval_env=1)
            for x in self._levels
        } if choice_formula else {}
        self._choice_x_design_info = {k: v.design_info
                                      for k, v in choice_predictors}

        # xdim = predictors.shape[1]
        # sample_size = predictors.shape[0]
        niter = int(niter)
        if niter <= 0:
            raise Exception("niter should be a positive integer.")

        if ping is None:
            ping = int(niter / 10)
        ping = int(ping)

        if seed is not None:
            boom.GlobalRng.rng.seed(int(seed))

        self._model = boom.MultinomialLogitModel(
            np.array(response, dtype="str"),
            R.to_boom_matrix(subject_predictors),
            [R.to_boom_matrix(choice_predictors[x] for x in levels)]
            if choice_predictors else [])

        if prior is None:
            prior = MultinomialLogitSpikeSlabPrior.from_model(self._model)
        prior.create_sampler(self._model, assign=True)

        # A lil matrix is a "linked list" matrix.  This is an efficient method
        # for constructing matrices.  It should be converted to a different
        # matrix type before doing anything with it.
        nvars = self._model.beta_size(include_zeros=False)
        self._coefficient_draws = scipy.sparse.lil_matrix((niter, nvars))
        self._log_likelihood = np.zeros(niter)

        for i in range(niter):
            self._model.sample_posterior()
            beta = self._model.coefficients
            self._coefficient_draws[i, :] = sparsify(beta)
            self._log_likelihood[i] = self._model.log_likelihood

        # Convert the coefficient draws to sparse column format.  Predictions
        # vs this format should take the form X @ beta, not beta @ X.
        self._coefficient_draws = self._coefficient_draws.tocsc()

    @property
    def xdim(self):
        return self._model.xdim

    @property
    def log_likelihood(self):
        return self._log_likelihood

    @property
    def xnames(self):
        # A list of strings containing the column names of the predictors.
        return self._x_design_info.column_names

    def predict(self, newdata, burn=None, seed=None):
        """
        Return an LmSpikePrediciton object.
        """
        if burn is None:
            burn = R.suggest_burn(self.log_likelihood)
        if seed is not None:
            boom.GlobalRng.rng.seed(int(seed))
        if isinstance(newdata, np.ndarray) and len(newdata.shape) == 1:
            newdata = newdata.reshape(1, -1)
        if isinstance(newdata, np.ndarray) and newdata.shape[1] == self.xdim:
            predictors = newdata
        else:
            predictors = patsy.build_design_matrices(
                [self._x_design_info],
                data=newdata)[0]
        return self._coefficient_draws[burn:, :] @ predictors.T

    @staticmethod
    def build_choice_formula(base_formula: str, levels, prefix="[X]"):
        """
        A common way to include choice-level predictors in a data frame is via a
        naming scheme like

        A_size, B_size, C_size, A_gas_mileage, B_gas_mileage, C_gas_mileage, ...

        This function allows you to specify a model formula along the lines of
        "[X]_size + [X]_gas_mileage".

        Args:
          base_formula: A string giving the model formula for the 'choice' part
            of the model in terms of a prefix that will be replaced by the
            levels of the response variable.
          levels:  A list containing the levels of the response variable.
          prefix: The dummy string to be used in 'base_formula' in place of the
            level names.

        Returns:
          A dict, keyed by the values in 'levels'.  Each dict entry is a string
            containing 'base_formula' but with 'prefix' replaced by the the
            values in 'levels'.
        """
        return {x: base_formula.replace(prefix, str(x)) for x in levels}



import BayesBoom.boom as boom
import numpy as np
import BayesBoom.R as R
from .state_models import StateModel


class SeasonalStateModel(StateModel):
    """
    Seasonal state model.  TODO(steve): build this.
    """

    def __init__(self,
                 y,
                 nseasons: int,
                 season_duration: int = 1,
                 initial_state_prior=None,
                 innovation_sd_prior: R.SdPrior = None,
                 sdy: float = None):
        """
        Args:
          y: The time series being modeled.  This can be omitted if either (a)
            initial_state_prior and sdy and initial_y are passed, or (b) sdy
            and initial_y are passed.
          nseasons: The number of seasons in a cycle.
          season_duration:  The number of time periods each season.  See below.
          initial_state_prior: A multivariate normal distribution of dimension
            nseasons - 1.  This is a distribution on the seasonal value at time
            0 and on the nseasons-2 previous values.  If None is passed then a
            default prior will be assumed.
          innovation_sd_prior: Prior distribution on the standard deviation of
            the innovation terms.  If None, then a default prior will be
            assumed.
          sdy: The standard deviation of the time series being modeled.

        Details:

        """
        self._nseasons = nseasons
        self._season_duration = season_duration

        if initial_state_prior is None:
            if sdy is None:
                if y is None:
                    raise Exception("One of 'y', 'sdy', or "
                                    "'initial_state_prior' must be supplied.")
                sdy = np.nanstd(y, ddof=1)
            initial_state_prior = self._default_initial_state_prior(sdy)

        if isinstance(initial_state_prior, R.NormalPrior):
            dim = nseasons - 1
            mu = initial_state_prior.mean
            sigma = initial_state_prior.sd
            initial_state_prior = R.MvnPrior(
                mu=np.fill(dim, mu),
                Sigma=np.diag(np.fill(dim, sigma * sigma)))

        if not isinstance(initial_state_prior, R.MvnPrior):
            raise Exception("Unexpected type for 'initial_state_prior'.  "
                            "Acceptable types include R.NormalPrior or "
                            "R.MvnPrior.")
        self._initial_state_prior = initial_state_prior

        if innovation_sd_prior is None:
            if sdy is None:
                if y is None:
                    raise Exception("One of 'y', 'sdy', or "
                                    "'innovation_sd_prior' must be supplied.")
                sdy = np.nanstd(y, ddof=1)
            innovation_sd_prior = self._default_sigma_prior(sdy)
        if not isinstance(innovation_sd_prior, R.SdPrior):
            raise Exception("Expected an R.SdPrior for innovation_sd_prior.")
        self._innovation_sd_prior = innovation_sd_prior

        self._build_model()
        self._state_contribution = None

    @property
    def label(self):
        ans = f"Seasonal[{self._nseasons}]"
        if self._season_duration > 1:
            ans += f"[{self._season_duration}]"
        return ans

    @property
    def nseasons(self):
        return self._nseasons

    @property
    def season_duration(self):
        return self._season_duration

    @property
    def state_dimension(self):
        return self.nseasons - 1

    @property
    def state_contribution(self):
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self.sigma_draws = np.zeros(niter)
        self._state_contribution = np.zeros((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self.sigma_draws[iteration] = self._state_model.sigma
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

    def restore_state(self, iteration):
        self._state_model.set_sigma(self.sigma_draws[iteration])

    def plot_state_contribution(
            self, fig, gridspec, time, burn=None, ylim=None, **kwargs):
        if self.nseasons == 7 and self.season_duration == 1:
            return self._plot_day_of_week_cycle(
                fig=fig, gridspec=gridspec, time=time, burn=burn,
                ylim=ylim, **kwargs)
        else:
            return self.plot_state_contribution_default(
                fig=fig, gridspec=gridspec, time=time, burn=burn,
                ylim=ylim, **kwargs)

    def _build_model(self):
        self._state_model = boom.SeasonalStateModel(
            nseasons=self._nseasons, season_duration=self._season_duration)

        self._state_model.set_initial_state_mean(
            boom.Vector(self._initial_state_prior.mu))
        self._state_model.set_initial_state_variance(
            boom.SpdMatrix(self._initial_state_prior.Sigma))
        self._state_model.set_sigma(self._innovation_sd_prior.initial_value)

        # The prior needs to be saved so the object can be serialized.
        self._assign_posterior_sampler(self._innovation_sd_prior)

    def __repr__(self):
        ans = f"A SeasonalStateModel with {self.nseasons} "
        ans += f"seasons of duration {self.season_duration}, and "
        ans += f"residual sd {self._state_model.sigma}."
        return ans

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_model()

    @staticmethod
    def _default_sigma_prior(sdy):
        """
        The default prior to use for the innovation standard deviation.
        """
        return R.SdPrior(.01 * sdy, upper_limit=sdy)

    def _assign_posterior_sampler(self, innovation_sd_prior: R.SdPrior):
        state_model_sampler = boom.ZeroMeanGaussianConjSampler(
            self._state_model,
            innovation_sd_prior.boom(),
            seeding_rng=boom.GlobalRng.rng)
        state_model_sampler.set_sigma_upper_limit(
            innovation_sd_prior.upper_limit)
        self._state_model.set_method(state_model_sampler)

    def _default_initial_state_prior(self, sdy):
        """
        The default prior to use for the initial state vector.
        """
        dim = self.nseasons - 1
        return R.MvnPrior(np.zeros(dim),
                          np.diag(np.full(dim, float(sdy))))

    def _plot_day_of_week_cycle(
            self, fig, gridspec, burn, time, ylim, **kwargs):
        spec = gridspec.subgridspec(3, 3)
        season = 0
        if burn is None:
            burn = 0
        for i in range(3):
            for j in range(3):
                if (j == 2) and (i < 2):
                    pass
                else:
                    if i == 0:
                        ax = fig.add_subplot(spec[i, j])
                        row_axis = ax
                    else:
                        ax = fig.add_subplot(spec[i, j], sharey=row_axis)
                    time_subset = time[season::7]
                    R.PlotDynamicDistribution(
                        curves=self.state_contribution[burn:, season::7],
                        timestamps=time_subset,
                        ax=ax,
                        ylim=ylim,
                        **kwargs)
                    ax.tick_params(
                        bottom=False,
                        top=False,
                        left=False,
                        right=False,
                        labelbottom=False,
                        labeltop=False,
                        labelleft=False,
                        labelright=False,
                    )
                    if (i == 2):
                        ax.tick_params(
                            bottom=True,
                            labelbottom=True)
                    elif (i == 0) and (j == 2):
                        ax.tick_params(
                            top=True,
                            labeltop=True)

                    if j == 0:
                        ax.tick_params(
                            left=True,
                            labelleft=True)
        return None




from .spikeslab import (
    BigAssSpikeSlab,
    compute_inclusion_probabilities,
    coefficient_positive_probability,
    dot,
    lm_spike,
    sparsify,
    set_glm_coefs,
    plot_inclusion_probs,
    plot_model_size,
)

from .mlogit_spike import mlogit_spike

from .priors import (
    LogitZellnerPrior,
    MultinomialLogitSpikeSlabPrior,
    RegressionSpikeSlabPrior,
    StudentSpikeSlabPrior,
)

__all__ = [
    "BigAssSpikeSlab",
    "LogitZellnerPrior",
    "MultinomialLogitSpikeSlabPrior",
    "RegressionSpikeSlabPrior",
    "StudentSpikeSlabPrior",
    "coefficient_positive_probability",
    "compute_inclusion_probabilities",
    "dot",
    "lm_spike",
    "mlogit_spike",
    "plot_inclusion_probs",
    "plot_model_size",
    "set_glm_coefs",
    "sparsify",
]



import matplotlib.pyplot as plt
import numpy as np

from .R import (
    data_frame,
    pretty,
    ls,
    table,
    data_range,
    corr,
    first_true,
    invert_order,
    order,
    paste,
    paste0,
    print_timestamp,
    recycle,
    remove_common_prefix,
    remove_common_suffix,
    unique_match,
    var,
    which,
)

from .bayes import (
    Ar1CoefficientPrior,
    BetaPrior,
    DoubleModel,
    GaussianSuf,
    MvnPrior,
    MvnGivenSigma,
    NormalPrior,
    SdPrior,
    UniformPrior,
    WishartPrior,
)

from .cbind import cbind

from .density import Density

from .data_table import to_data_table, to_data_frame

from .autoclean import AutoClean

from .encoding import (
    register_encoding_json_encoder,
    Encoder,
    MainEffectEncoder,
    MainEffectEncoderJsonEncoder,
    MainEffectEncoderJsonDecoder,
    EffectEncoder,
    OneHotEncoder,
    IdentityEncoder,
    InteractionEncoder,
    SuccessEncoder,
    SuccessEncoderJsonEncoder,
    SuccessEncoderJsonDecoder,
    DatasetEncoder,
    DatasetEncoderJsonEncoder,
    DatasetEncoderJsonDecoder,
)

from .graphics_device import get_current_graphics_device

from .mcmc import suggest_burn, report_progress

from .plots import (
    abline,
    barplot,
    boxplot,
    compare_dynamic_distributions,
    hist,
    histabunch,
    hosmer_lemeshow_plot,
    lines,
    lty,
    mosaic_plot,
    pairs,
    plot,
    plot_dynamic_distribution,
    PlotDynamicDistribution,
    plot_many_ts,
    plot_grid_shape,
    plot_ts,
    points,
    time_series_boxplot,
    )

from .probability import (
    dmvn, rmvn,
    dnorm, pnorm, qnorm, rnorm,
    dgamma, pgamma, qgamma, rgamma,
    rbeta,
    rpois,
    rmarkov,
)

from .stats import density, sd

from .test_utilities import delete_if_present

from .boom_py_utils import (
    is_all_numeric,
    is_iterable,
    to_boom_date,
    to_boom_vector,
    to_boom_matrix,
    to_boom_spd,
    to_numpy,
    to_pd_timestamp,
)
